"""A parser for Comet output"""
import re
import logging

import pandas as pd

from .txt import read_txt
from .. import utils

LOGGER = logging.getLogger(__name__)


def read_comet(
    txt_files, pairing_file_name=None, decoy_prefix="DECOY_", copy_data=True
):
    """Read peptide-spectrum matches (PSMs) from Comet output.
    Can parse tab-delimited files.

    Parameters
    ----------
    txt_files : str, pandas.DataFrame or tuple of str
        One or more collection of PSMs in the Crux tab-delimited format.
    pairing_file_name : str, optional
        A tab-delimited file that explicity pairs target and decoy peptide
        sequences. Requires one column labled 'target' that contains target
        sequences and a second colun labeled 'decoy' that contains decoy
        sequences. This file can be generated by setting --peptide-list=T
        in tide-index.
    decoy_prefix : str, optional
        The prefix used to indicate a decoy protein in the protein column.
        Default value is 'DECOY_'.
    copy_data : bool, optional
        If true, a deep copy of the data is created. This uses more memory, but
        is safer because it prevents accidental modification of the underlying
        data. This argument only has an effect when `txt_files` is a
        :py:class:`pandas.DataFrame`

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object containing the parsed
        PSMs.
    """
    target = "target/decoy"
    peptide = "modified_peptide"
    spectrum = ["scan", "exp_neutral_mass"]  # No file col to use
    pairing = ""
    protein = "protein"
    protein_delim = ","

    # Possible score columns output by Comet.
    scores = {
        "e-value",
        "xcorr",
        "sp_score",
    }
    scores_all = scores

    # Keep only Comet scores that exist in all of the files.
    skip_first_line = False
    crux_comet = False  # tracks whether standalone comet or Crux Comet
    if isinstance(txt_files, pd.DataFrame):
        scores = scores.intersection(set(txt_files.columns))
    else:
        txt_files = utils.listify(txt_files)
        for txt_file in txt_files:
            with open(txt_file) as txt_ref:
                # First line of Comet output consists only of version
                # If statement below in case first line is removed
                line = txt_ref.readline().rstrip()
                if line.startswith("CometVersion"):
                    line = txt_ref.readline().rstrip()
                    skip_first_line = True
                cols = line.split("\t")
                scores = scores.intersection(set(cols))

                # Standalone Comet and Crux Comet use different column headers
                if protein not in cols:
                    protein = "protein id"
                    peptide = "modified sequence"
                    spectrum = ["scan", "spectrum neutral mass"]
                    crux_comet = True

    if not scores:
        raise ValueError(
            "Could not find any of the Comet score columns in all of the files."
            f" The columns Crema looks for are {', '.join(list(scores_all))}"
        )

    scores = list(scores)

    # Read in the files:
    fields = [*spectrum, peptide, target, *scores, pairing, protein]
    if isinstance(txt_files, pd.DataFrame):
        data = txt_files.copy(deep=copy_data).loc[:, fields]
    else:
        data = pd.concat(
            [
                utils.parse_psms_txt(f, fields, skip_first_line)
                for f in txt_files
            ]
        )

    if crux_comet and decoy_prefix == "DECOY_":
        decoy_prefix = "decoy_"
    data["target/decoy"] = ~data[protein].str.contains(decoy_prefix)

    psms = read_txt(
        data,
        target_column=target,
        spectrum_columns=spectrum,
        score_columns=scores,
        peptide_column=peptide,
        protein_column=protein,
        protein_delim=protein_delim,
        sep="\t",
        copy_data=False,
    )

    # Remove first and last amino acid from sequence
    # Looks like "R.WVNEK.Y"
    peptide_column = psms.peptides
    new_peptide_column = peptide_column.str[2:-2]
    psms.set_peptide_column(new_peptide_column)

    # always pair target and decoys for Comet
    if pairing_file_name == None:
        # implicit pairing based off fact that Comet reverses peptides
        psms._peptide_pairing = _create_pairing(
            data, peptide, protein, decoy_prefix
        )
    else:  # explicit pairing
        psms._peptide_pairing = utils.create_pairing_from_file(
            pairing_file_name
        )

    # Remove decoy prefix from protein ID
    protein_column = psms.proteins
    new_protein_column = protein_column.str.replace(
        decoy_prefix, "", regex=True
    )
    psms.set_protein_column(new_protein_column)

    return psms


def _create_pairing(pairing_data, peptide_col, protein_col, decoy_prefix):
    """Parse a single Comet dataframe to implicity pair target and
    decoy sequences.

    Parameters
    ----------
    pairing_data : pandas.DataFrame
        A collection of PSMs with the necessary columns to create a
        target/decoy peptide pairing. Required columns are modified sequence
        and protein ID.
    peptide_col: str
        Identifies the modified sequence column.
    protein_col: str
        Identifies the protein ID column.
    decoy_prefix: str
        The prefix used to indicate a decoy protein in the protein column.

    Returns
    -------
    pairing : dict
        A map of target and decoy peptide sequence pairings. Targets with
        missing decoys will not be included among the keys.

    """
    # ensure pairing_data dataframe contains all necessary columns
    req_fields = [peptide_col, protein_col]

    if not set(req_fields).issubset(pairing_data.columns):
        miss = ", ".join(set(req_fields) - set(pairing_data.columns))
        raise ValueError(
            f"Required columns for peptide pairing were not detected: {miss}"
        )

    pairing_data = pairing_data.loc[:, req_fields]
    pairing_data[peptide_col] = pairing_data[peptide_col].str[2:-2]

    pairing_data["target/decoy"] = ~pairing_data[protein_col].str.contains(
        decoy_prefix
    )

    reverse_peptide_list = []
    for seq in list(pairing_data[peptide_col]):
        seq_sp = re.split(r"(?<=.)(?=[A-Z])", seq)
        peptide_rev = "".join([seq_sp[0], *reversed(seq_sp[1:-1]), seq_sp[-1]])
        reverse_peptide_list.append(peptide_rev)

    pairing_data["reverse_peptide"] = reverse_peptide_list

    targets = pairing_data[pairing_data["target/decoy"]]
    decoys = pairing_data[~pairing_data["target/decoy"]]

    dic1 = dict(zip(targets[peptide_col], targets["reverse_peptide"]))
    dic2 = dict(zip(decoys["reverse_peptide"], decoys[peptide_col]))

    dic2.update(dic1)
    return dic2



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import subprocess
import crema

sys.path.insert(0, os.path.abspath(os.path.join("..", "..")))

try:
    import numpydoc
except ModuleNotFoundError:
    subprocess.run(["pip", "install", "numpydoc"], check=True)


# -- Project information -----------------------------------------------------
project = "crema"
copyright = "2020, Donavan See"
author = "Andy Lin, William E. Fondrie, Donavan See, William S. Noble"
version = str(crema.__version__)
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "numpydoc",
    "sphinx_rtd_theme",
    "sphinx.ext.intersphinx",
    "sphinxarg.ext",
    "nbsphinx",
]

nbsphinx_execute = "never"

autosummary_generate = True
autodoc_default_options = {
    "members": True,
    "inherited-members": True,
    "member-order": "groupwise",
}

numpydoc_show_class_members = True
numpydoc_show_inherited_class_members = True
numpydoc_attributes_as_param_list = True
intersphinx_mapping = {
    "pandas": ("https://pandas.pydata.org/docs", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "numpy": ("https://docs.scipy.org/doc/numpy", None),
    "matplotlib": ("https://matplotlib.org", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
# html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

master_doc = "index"
source_suffix = ".rst"

html_theme_options = {
    "style_nav_header_background": "#343131",
    "logo_only": True,
}

html_css_files = [
    "custom.css",
]

html_logo = "_static/crema_logo_caramel_light.png"



"""The :py:class:`PsmDataset` class is used to define a collection of
peptide-spectrum matches.
"""
import logging

from .confidence import TdcConfidence
from .confidence import MixmaxConfidence
from .qvalues import tdc
from .utils import listify

LOGGER = logging.getLogger(__name__)


class PsmDataset:
    """Store a collection of peptide-spectrum matches (PSMs).

    Parameters
    ----------
    psms : pandas.DataFrame
        A :py:class:`pandas.DataFrame` of PSMs.
    target_column : str
        The column that indicates whether a PSM is a target or a decoy. This
        column should be boolean, where :code:`True` indicates a target and
        :code:`False` indicates a decoy.
    spectrum_columns : str or tuple of str
        One or more columns that together define a unique mass spectrum.
    score_columns : str or tuple of str, optional
        One or more columns that indicate scores by which crema can rank PSMs.
    peptide_column : str
        The column that defines a unique peptide. Modifications should be
        indicated either in square brackets :code:`[]` or parentheses
        :code:`()`. The exact modification format within these entities does
        not matter, so long as it is consistent.
    protein_columns : str
        The column that defines a unique protein.
    protein_delim : str
        The string delimiter that is needed to separate multiple proteins found
        in the protein column.
    peptide_pairing: dict
        A map of target and decoy peptide sequence pairings to be used for TDC.
        This should be in the form {key=target_sequence:value=decoy_sequence}
        where decoy sequences are shuffled versions of target sequences.
    copy_data : bool, optional
        If true, a deep copy of the data is created. This uses more memory, but
        is safer because it prevents accidental modification of the underlying
        data. This argument only has an effect when `pin_files` is a
        :py:class:`pandas.DataFrame`

    Attributes
    ----------
    data : pandas.DataFrame
    spectrum_columns : list of str
    score_columns : list of str
    target_column : str
    peptide_column : str
    protein_column : str
    protein_delim : str
    methods : dict
    peptide_pairing : dict
    """

    methods = {"tdc": TdcConfidence, "mixmax": MixmaxConfidence}

    def __init__(
        self,
        psms,
        target_column,
        spectrum_columns,
        score_columns,
        peptide_column,
        protein_column,
        protein_delim,
        peptide_pairing=None,
        copy_data=True,
    ):
        """Initialize a PsmDataset object."""
        self.score_columns = listify(score_columns)
        self._spectrum_columns = listify(spectrum_columns)
        self._target_column = target_column
        self._peptide_column = peptide_column
        self._protein_column = protein_column
        self._protein_delim = protein_delim
        self._peptide_pairing = peptide_pairing

        fields = sum(
            [
                self._spectrum_columns,
                self.score_columns,
                [self._target_column],
                [self._peptide_column],
                [self._protein_column],
            ],
            [],
        )
        self._data = psms.copy(deep=copy_data).loc[:, fields]
        self._data[target_column] = self._data[target_column].astype(bool)
        self._num_targets = self.targets.sum()
        self._num_decoys = (~self.targets).sum()

        if self.data.empty:
            raise ValueError("No PSMs were detected.")

        if not self._num_decoys:
            raise ValueError("No decoy PSMs were detected.")

        if not self._num_targets:
            raise ValueError("No target PSMs were detected.")

    @property
    def columns(self):
        """The columns of the PSM :py:class:`pandas.DataFrame`"""
        return self._data.columns

    @property
    def data(self):
        """The collection of PSMs as a :py:class:`pandas.DataFrame`."""
        return self._data.copy()

    @property
    def spectra(self):
        """The mass spectrum identifiers as a :py:class:`pandas.DataFrame`."""
        return self[self._spectrum_columns]

    @property
    def peptides(self):
        """The peptides as a :py:class:`pandas.Series`."""
        return self[self._peptide_column]

    @property
    def proteins(self):
        """The proteins as a :py:class:`pandas.Series`."""
        return self[self._protein_column]

    @property
    def protein_delim(self):
        """The delimiter to split protein IDs as a string."""
        return self[self._protein_delim]

    @property
    def scores(self):
        """The scores for each PSM as a :py:class:`pandas.DataFrame`."""
        return self[self.score_columns]

    @property
    def targets(self):
        """An array indicating whether each PSM is a target"""
        return self[self._target_column].values

    @property
    def peptide_pairing(self):
        """A dictionary containing target/decoy peptide pairs"""
        return self._peptide_pairing

    def __getitem__(self, column):
        """Return the specified column"""
        return self._data.loc[:, column]

    def assign_confidence(
        self,
        score_column=None,
        desc=None,
        eval_fdr=0.01,
        method="tdc",
        pep_fdr_type="psm-peptide",
        prot_fdr_type="best",
        threshold=0.01,
    ):
        """Assign confidence estimates to this collection of peptide-spectrum matches.

        Parameters
        ----------
        score_column : str, optional
            The score by which to rank the PSMs for confidence estimation. If
            :code:`None`, the score that yields the most PSMs at the specified
            false discovery rate threshold (`eval_fdr`), will be used.
        desc : bool, optional
            True if higher scores better, False if lower scores are better.
            If None, crema will try both and use the
            choice that yields the most PSMs at the specified false discovery
            rate threshold (`eval_fdr`). If `score_column` is :code:`None`,
            this parameter is ignored.
        eval_fdr : float, optional
            The false discovery rate threshold used to evaluate the best
            `score_column` and `desc` to choose. This should range from 0 to 1.
        method : {"tdc"}, optional
            The method for crema to use when calculating the confidence estimates.
        pep_fdr_type : {"psm-only","peptide-only",psm-peptide"}, optional
            The method for crema to use when calculating peptide level confidence
            estimates. Default is "psm-peptide".
        prot_fdr_type : {"best", "combine"}, optional
            The method for crema to use when calculating protein level confidence
            estimates. Default is "best".
        threshold : float or "q-value", optional
            The FDR threshold for accepting discoveries. Default is 0.01. If
            "q-value" is chosen, then "accept" column is replaced with
            "crema q-value".

        Returns
        -------
        Confidence object
            The confidence estimates for this PsmDataset.
        """
        if score_column is None:
            score_column, _, desc = self.find_best_score(eval_fdr)

        conf = self.methods[method](
            psms=self,
            score_column=score_column,
            desc=desc,
            eval_fdr=eval_fdr,
            pep_fdr_type=pep_fdr_type,
            prot_fdr_type=prot_fdr_type,
            threshold=threshold,
        )

        return conf

    def find_best_score(self, eval_fdr=0.01):
        """Find the best score for this collection of PSMs

        Try each of the available score columns, determining how many PSMs
        are detected below the provided false discovery rate threshold. The
        best score is the one that returns the most.

        Parameters
        ----------
        eval fdr : float
            The false discovery rate threshold used to find the best score.

        Returns
        -------
        best_score : str
            The best score.
        n_passing : int
            The number of PSMs that meet the specified FDR threshold.
        desc : bool
            True if higher scores better, False if lower scores are better.
        """
        best_score = None
        best_passing = 0
        for desc in (True, False):
            qvals = self.scores.apply(tdc, target=self.targets, desc=desc)
            num_passing = (qvals <= eval_fdr).sum()
            feat_idx = num_passing.idxmax()
            num_passing = num_passing[feat_idx]
            if num_passing > best_passing:
                best_passing = num_passing
                best_score = feat_idx
                best_desc = desc

        if best_score is None:
            raise RuntimeError("No PSMs were found below the 'eval_fdr'.")

        return best_score, best_passing, best_desc

    def set_protein_column(self, new_protein_column):
        """Replaces current protein column with input protein column

        Parameters
        ----------
        new_protein_column : pandas.Series

        Returns
        -------

        """
        self._data[self._protein_column] = new_protein_column
        return

    def set_peptide_column(self, new_peptide_column):
        """Replaces current peptide column with input peptide column

        Parameters
        ----------
        new_peptide_column : pandas.Series

        Returns
        -------

        """
        self._data[self._peptide_column] = new_peptide_column
        return



"""These tests verify that confidence estimates are correctly produced via tdc through various parsers"""
import unittest

import numpy as np
import pandas as pd
from crema.parsers.tide import read_tide


def test_tide_tdc(target_tide_txt, decoy_tide_txt, tmp_path):
    expected_peptide_pairing = {
        "APPLE": "APLPE",
        "BANANA": "ANANAB",
        "CHERRY": "CHRREY",
        "DURIAN": "DRIUAN",
        "EGGPLANT": "EGPGLATN",
    }
    expected_target_psms = pd.DataFrame(
        [
            ["f1", 3, "CHERRY", "p3", 0.1, 1 / 5],
            ["f1", 7, "BANANA", "p2", 0.2, 1 / 5],
            ["f1", 5, "EGGPLANT", "p5", 0.25, 1 / 5],
            ["f1", 10, "EGGPLANT", "p5", 0.3, 1 / 5],
            ["f1", 2, "BANANA", "p2", 0.4, 1 / 5],
            ["f1", 1, "APPLE", "p1", 0.5, 2 / 9],
            ["f1", 4, "DURIAN", "p4", 0.55, 2 / 9],
            ["f1", 6, "APPLE", "p1", 0.60, 2 / 9],
            ["f1", 8, "CHERRY", "p3", 0.70, 2 / 9],
        ],
        columns=[
            "file",
            "scan",
            "sequence",
            "protein id",
            "combined p-value",
            "crema q-value",
        ],
    )
    expected_decoy_psms = pd.DataFrame(
        [
            ["f1", 9, "DRIUAN", "p4", 0.5],
        ],
        columns=[
            "file",
            "scan",
            "sequence",
            "protein id",
            "combined p-value",
        ],
    )
    expected_target_peptides = pd.DataFrame(
        [
            ["f1", 3, "CHERRY", "p3", 0.1, 1 / 3],
            ["f1", 7, "BANANA", "p2", 0.2, 1 / 3],
            ["f1", 5, "EGGPLANT", "p5", 0.25, 1 / 3],
            ["f1", 1, "APPLE", "p1", 0.5, 1 / 2],
        ],
        columns=[
            "file",
            "scan",
            "sequence",
            "protein id",
            "combined p-value",
            "crema q-value",
        ],
    )
    expected_decoy_peptides = pd.DataFrame(
        [
            ["f1", 9, "DRIUAN", "p4", 0.5],
        ],
        columns=[
            "file",
            "scan",
            "sequence",
            "protein id",
            "combined p-value",
        ],
    )
    expected_target_proteins = pd.DataFrame(
        [
            ["p3", 0.1, 1 / 3],
            ["p2", 0.2, 1 / 3],
            ["p5", 0.25, 1 / 3],
            ["p1", 0.5, 2 / 4],
        ],
        columns=[
            "protein id",
            "combined p-value",
            "crema q-value",
        ],
    )
    expected_decoy_proteins = pd.DataFrame(
        [
            ["p4", 0.5],
        ],
        columns=[
            "protein id",
            "combined p-value",
        ],
    )

    psms = read_tide([target_tide_txt, decoy_tide_txt])
    conf = psms.assign_confidence(
        score_column="combined p-value", desc=False, threshold="q-value"
    )

    unittest.TestCase().assertDictEqual(
        expected_peptide_pairing, psms.peptide_pairing
    )
    np.testing.assert_array_equal(
        expected_target_psms.values,
        conf.confidence_estimates["psms"].values,
    )
    np.testing.assert_array_equal(
        expected_decoy_psms.values,
        conf.decoy_confidence_estimates["psms"].values,
    )
    np.testing.assert_array_equal(
        expected_target_peptides.values,
        conf.confidence_estimates["peptides"].values,
    )
    np.testing.assert_array_equal(
        expected_decoy_peptides.values,
        conf.decoy_confidence_estimates["peptides"].values,
    )
    np.testing.assert_array_equal(
        expected_target_proteins.values,
        conf.confidence_estimates["proteins"].values,
    )
    np.testing.assert_array_equal(
        expected_decoy_proteins.values,
        conf.decoy_confidence_estimates["proteins"].values,
    )


def test_tide_tdc_prot_combine(target_tide_txt, decoy_tide_txt, tmp_path):
    # PSM and peptide level test are done above in function test_tide_tdc
    # This test is specifically to test the "combine" version of protein level
    # FDR
    psms = read_tide([target_tide_txt, decoy_tide_txt])
    conf = psms.assign_confidence(
        score_column="combined p-value",
        desc=False,
        prot_fdr_type="combine",
        threshold="q-value",
    )

    expected_target_proteins = pd.DataFrame(
        [
            ["p3", 0.1 * 0.7, 1 / 4],
            ["p5", 0.25 * 0.3, 1 / 4],
            ["p2", 0.4 * 0.2, 1 / 4],
            ["p1", 0.5 * 0.6, 1 / 4],
        ],
        columns=[
            "protein id",
            "combined p-value",
            "crema q-value",
        ],
    )
    expected_decoy_proteins = pd.DataFrame(
        [
            ["p4", 0.5],
        ],
        columns=[
            "protein id",
            "combined p-value",
        ],
    )

    np.testing.assert_array_equal(
        expected_target_proteins.values,
        conf.confidence_estimates["proteins"].values,
    )
    np.testing.assert_array_equal(
        expected_decoy_proteins.values,
        conf.decoy_confidence_estimates["proteins"].values,
    )



"""Fixtures that are used in multiple tests"""
from pathlib import Path

import pytest
import pandas as pd


@pytest.fixture
def basic_tide_df():
    """A simple tide-like dataframe"""
    df = pd.DataFrame(
        [
            ["f1", 1, 10, 0.7, "APPLE", "target", 0.7, "p1", "APPLE"],
            ["f1", 2, 20, 0.4, "ANANAB", "decoy", 0.1, "p2", "BANANA"],
            ["f1", 3, 30, 0.1, "CHERRY", "target", 0.2, "p3", "CHERRY"],
            ["f1", 4, 40, 0.55, "DURIAN", "target", 0.8, "p4", "DURIAN"],
            ["f1", 5, 50, 0.25, "EGGPLANT", "target", 0.25, "p5", "EGGPLANT"],
            ["f1", 1, 10, 0.6, "FIG", "target", 0.6, "p6", "FIG"],
            ["f1", 2, 20, 0.2, "GARPE", "decoy", 0.2, "p7", "GRAPE"],
            ["f1", 3, 30, 0.7, "HEONDEYW", "decoy", 0.4, "p8", "HONEYDEW"],
            ["f1", 4, 40, 0.56, "ICE", "target", 0.56, "p9", "ICE"],
            ["f1", 5, 50, 0.3, "JMA", "decoy", 0.3, "p10", "JAM"],
        ],
        columns=[
            "file",
            "scan",
            "spectrum precursor m/z",
            "combined p-value",
            "sequence",
            "target/decoy",
            "x",
            "protein id",
            "original target sequence",
        ],
    )
    return df


@pytest.fixture
def target_tide_df():
    """A tide-like dataframe of target psms"""
    df = pd.DataFrame(
        [
            ["f1", 1, 10, 101, 0.5, "APPLE", "target", "p1"],
            ["f1", 2, 20, 102, 0.4, "BANANA", "target", "p2"],
            ["f1", 3, 30, 103, 0.1, "CHERRY", "target", "p3"],
            ["f1", 4, 40, 104, 0.55, "DURIAN", "target", "p4"],
            ["f1", 5, 50, 105, 0.25, "EGGPLANT", "target", "p5"],
            ["f1", 6, 60, 106, 0.6, "APPLE", "target", "p1"],
            ["f1", 7, 70, 107, 0.2, "BANANA", "target", "p2"],
            ["f1", 8, 80, 108, 0.7, "CHERRY", "target", "p3"],
            ["f1", 9, 90, 109, 0.56, "DURIAN", "target", "p4"],
            ["f1", 10, 100, 110, 0.3, "EGGPLANT", "target", "p5"],
        ],
        columns=[
            "file",
            "scan",
            "spectrum precursor m/z",
            "peptide mass",
            "combined p-value",
            "sequence",
            "target/decoy",
            "protein id",
        ],
    )
    return df


@pytest.fixture
def decoy_tide_df():
    """A tide-like dataframe of decoy psms"""
    df = pd.DataFrame(
        [
            ["f1", 1, 10, 101, 0.7, "APLPE", "decoy", "APPLE", "p1"],
            ["f1", 2, 20, 102, 0.45, "ANANAB", "decoy", "BANANA", "p2"],
            ["f1", 3, 30, 103, 0.3, "CHRREY", "decoy", "CHERRY", "p3"],
            ["f1", 4, 40, 104, 0.6, "DRIUAN", "decoy", "DURIAN", "p4"],
            ["f1", 5, 50, 105, 0.9, "EGPGLATN", "decoy", "EGGPLANT", "p5"],
            ["f1", 6, 60, 106, 0.75, "APLPE", "decoy", "APPLE", "p1"],
            ["f1", 7, 70, 107, 0.25, "ANANAB", "decoy", "BANANA", "p2"],
            ["f1", 8, 80, 108, 0.9, "CHRREY", "decoy", "CHERRY", "p3"],
            ["f1", 9, 90, 109, 0.5, "DRIUAN", "decoy", "DURIAN", "p4"],
            ["f1", 10, 100, 110, 0.8, "EGPGLATN", "decoy", "EGGPLANT", "p5"],
        ],
        columns=[
            "file",
            "scan",
            "spectrum precursor m/z",
            "peptide mass",
            "combined p-value",
            "sequence",
            "target/decoy",
            "original target sequence",
            "protein id",
        ],
    )
    return df


@pytest.fixture
def mod_target_tide_df():
    """A tide-like dataframe of target psms"""
    df = pd.DataFrame(
        [
            ["f1", 1, 10, 101, 0.1, "ALLSLR", "target", "p1"],
            ["f1", 2, 20, 102, 0.2, "AILSIR", "target", "p2"],
            ["f1", 3, 30, 103, 0.3, "LAVITR", "target", "p3"],
            ["f1", 4, 40, 104, 0.4, "QTPPAR", "target", "p4"],
            ["f1", 5, 50, 105, 0.5, "IPLVNL", "target", "p5"],
            ["f1", 6, 60, 106, 0.6, "SRGPPR", "target", "p6"],
            ["f1", 7, 70, 107, 0.7, "GEVPN[0.98]R", "target", "p7"],
            ["f1", 8, 80, 108, 0.8, "GGHMDR", "target", "p8"],
            ["f1", 9, 90, 109, 0.9, "NPANRT", "target", "p9"],
            ["f1", 10, 100, 110, 0.99, "SADAGPR", "target", "p10"],
        ],
        columns=[
            "file",
            "scan",
            "spectrum precursor m/z",
            "peptide mass",
            "combined p-value",
            "sequence",
            "target/decoy",
            "protein id",
        ],
    )
    return df


@pytest.fixture
def mod_decoy_tide_df():
    """A tide-like dataframe of target psms with peptide modifications"""
    df = pd.DataFrame(
        [
            ["f1", 1, 10, 101, 0.1, "ALLLSR", "decoy", "ALLSLR", "p1"],
            ["f1", 2, 20, 102, 0.3, "ALLISR", "decoy", "ASLILR", "p2"],
            ["f1", 3, 30, 103, 0.2, "LATVIR", "decoy", "LIVATR", "p3"],
            ["f1", 4, 40, 104, 0.6, "QTAPPR", "decoy", "QTPPAR", "p4"],
            ["f1", 5, 50, 105, 0.5, "IVLN[0.98]PL", "decoy", "IPLVNL", "p5"],
            ["f1", 6, 60, 106, 0.9, "RGSPPR", "decoy", "RSGPPR", "p6"],
            ["f1", 7, 70, 107, 0.4, "GN[0.98]PEVR", "decoy", "GEVPNR", "p7"],
            ["f1", 8, 80, 108, 0.7, "GDMGHR", "decoy", "GGHMDR", "p8"],
            ["f1", 9, 90, 109, 0.8, "N[0.98]NPTAR", "decoy", "NPANTR", "p9"],
            ["f1", 10, 100, 110, 0.55, "SGDAPAR", "decoy", "SAADGPR", "p10"],
        ],
        columns=[
            "file",
            "scan",
            "spectrum precursor m/z",
            "peptide mass",
            "combined p-value",
            "sequence",
            "target/decoy",
            "original target sequence",
            "protein id",
        ],
    )
    return df


@pytest.fixture
def basic_comet_df():
    """A simple comet-like dataframe"""
    df = pd.DataFrame(
        [
            [1, 10, 0.7, "K.A[15.9949]PPLE.A", "prot1"],
            [2, 20, 0.4, "K.BNANAA.A", "DECOY_prot2"],
            [3, 30, 0.1, "K.CH[15.9949]E[15.9949]RRY.A", "prot3"],
            [4, 40, 0.55, "K.DURIAN.A", "prot4"],
            [5, 50, 0.25, "K.EGGPLANT.A", "prot5"],
            [1, 10, 0.6, "K.FIGS.A", "prot6"],
            [2, 20, 0.2, "K.GPARE.A", "DECOY_prot7"],
            [3, 30, 0.7, "K.H[15.9949]E[15.9949]DYENOW.A", "DECOY_prot8"],
            [4, 40, 0.56, "K.IC[15.9949]ES.A", "prot9"],
            [5, 50, 0.3, "K.JMAS.A", "DECOY_prot10"],
        ],
        columns=[
            "scan",
            "exp_neutral_mass",
            "e-value",
            "modified_peptide",
            "protein",
        ],
    )
    return df


@pytest.fixture
def basic_msgf_df():
    """A simple MSGF+-like dataframe"""
    df = pd.DataFrame(
        [
            ["f1", 0, 0, "HCD", 1, 0, 1, 2, "ANT", "p1", 1, 1, 1, 1, 1, 1],
            ["f1", 1, 1, "HCD", 1, 0, 1, 2, "ANT", "p1", 1, 1, 1, 1, 1, 1],
            ["f1", 2, 2, "HCD", 1, 0, 1, 2, "CAT", "XXX_p1", 1, 1, 1, 1, 1, 1],
            ["f1", 3, 3, "HCD", 1, 0, 1, 2, "DAB", "p2", 1, 1, 1, 1, 1, 1],
            ["f1", 4, 4, "HCD", 1, 0, 1, 2, "EVE", "p3", 1, 1, 1, 1, 1, 1],
            ["f1", 5, 5, "HCD", 1, 0, 1, 2, "FOE", "XXX_p3", 1, 1, 1, 1, 1, 1],
            ["f1", 6, 6, "HCD", 1, 0, 1, 2, "HEY", "p4", 1, 1, 1, 1, 1, 1],
            ["f1", 7, 7, "HCD", 1, 0, 1, 2, "INC", "p5", 1, 1, 1, 1, 1, 1],
            ["f1", 8, 8, "HCD", 1, 0, 1, 2, "JET", "XXX_p6", 1, 1, 1, 1, 1, 1],
            ["f1", 9, 9, "HCD", 1, 0, 1, 2, "KID", "p7", 1, 1, 1, 1, 1, 1],
        ],
        columns=[
            "#SpecFile",
            "SpecID",
            "ScanNum",
            "FragMethod",
            "Precursor",
            "IsotopeError",
            "PrecursorError(ppm)",
            "Charge",
            "Peptide",
            "Protein",
            "DeNovoScore",
            "MSGFScore",
            "SpecEValue",
            "EValue",
            "QValue",
            "PepQValue",
        ],
    )
    return df


@pytest.fixture
def basic_msamanda_df():
    """A simple MSAmanda-like dataframe"""
    df = pd.DataFrame(
        [
            [0, "X", "APPLE", "", "p1", 15, 0.1, 1, 1, 1, 2, 3, "f1"],
            [1, "X", "APPLE", "", "p1", 21, 0.2, 1, 1, 1, 2, 3, "f1"],
            [2, "X", "BAT", "", "p3", 35, 0.3, 1, 1, 1, 2, 3, "f1"],
            [3, "X", "SAT", "", "p4", 87, 0.8, 1, 1, 1, 2, 3, "f1"],
            [4, "X", "ELPPA", "", "REV_p1", 10, 0.1, 1, 1, 1, 2, 3, "f1"],
            [5, "X", "APPLE", "", "REV_p2", 40, 0.5, 1, 1, 1, 2, 3, "f1"],
            [6, "X", "APPLE", "", "REV_p1", 75, 0.7, 1, 1, 1, 2, 3, "f1"],
            [7, "X", "TAB", "", "REV_p3", 90, 0.7, 1, 1, 1, 2, 3, "f1"],
            [8, "X", "LAP", "", "REV_p4", 10, 0.2, 1, 1, 1, 2, 3, "f1"],
            [9, "X", "JACK", "", "REV_p5", 30, 0.3, 1, 1, 1, 2, 3, "f1"],
        ],
        columns=[
            "Scan Number",
            "Title",
            "Sequence",
            "Modifications",
            "Protein Accessions",
            "Amanda Score",
            "Weighted Probability",
            "Rank",
            "m/z",
            "Charge",
            "RT",
            "Nr of matched peaks",
            "Filename",
        ],
    )
    return df


@pytest.fixture
def basic_tide_txt(basic_tide_df, tmp_path):
    """A simple tide-like txt file"""
    out_file = tmp_path / "tide.txt"
    basic_tide_df.to_csv(out_file, sep="\t", index=False)
    return out_file


@pytest.fixture
def target_tide_txt(target_tide_df, tmp_path):
    """A tide-like txt file of target psms"""
    out_file = tmp_path / "target_tide.txt"
    target_tide_df.to_csv(out_file, sep="\t", index=False)
    return out_file


@pytest.fixture
def decoy_tide_txt(decoy_tide_df, tmp_path):
    """A tide-like txt file of decoy psms"""
    out_file = tmp_path / "decoy_tide.txt"
    decoy_tide_df.to_csv(out_file, sep="\t", index=False)
    return out_file


@pytest.fixture
def mod_target_tide_txt(mod_target_tide_df, tmp_path):
    """A tide-like txt file of target psms"""
    out_file = tmp_path / "mod_target_tide.txt"
    mod_target_tide_df.to_csv(out_file, sep="\t", index=False)
    return out_file


@pytest.fixture
def mod_decoy_tide_txt(mod_decoy_tide_df, tmp_path):
    """A tide-like txt file of decoy psms"""
    out_file = tmp_path / "mod_decoy_tide.txt"
    mod_decoy_tide_df.to_csv(out_file, sep="\t", index=False)
    return out_file


@pytest.fixture
def basic_tide_csv(basic_tide_df, tmp_path):
    """A simple tide-like csv file"""
    out_file = tmp_path / "tide.csv"
    basic_tide_df.to_csv(out_file, sep=",", index=False)
    return out_file


@pytest.fixture
def basic_msamanda_txt(basic_msamanda_df, tmp_path):
    """A simple MSAmanda-like txt file"""
    out_file = tmp_path / "msamanda.txt"
    print(tmp_path)
    with open(out_file, "w") as tmp_file:
        tmp_file.write("#version: 2.0.0.18350\n")
    basic_msamanda_df.to_csv(out_file, mode="a", sep="\t", index=False)
    return out_file


@pytest.fixture
def basic_msgf_txt(basic_msgf_df, tmp_path):
    "A simple MSGF+-liek tsv file" ""
    out_file = tmp_path / "msgf.txt"
    basic_msgf_df.to_csv(out_file, sep="\t", index=False)
    return out_file


@pytest.fixture
def real_tide_txt():
    """Return real tide txt files"""
    targets = Path("data/example_psms_target.txt")
    decoys = Path("data/example_psms_decoy.txt")
    return [targets, decoys]


@pytest.fixture
def real_mztab():
    """Return a real mzTab file"""
    return Path("data/MSV000085729.mzTab")


@pytest.fixture
def real_pepxml():
    """Return a real pepXML file"""
    return Path("data/tide-search.pep.xml")


@pytest.fixture
def real_msfragger_pepxml():
    """Return a real MSFragger file"""
    return Path("data/msfragger.pepxml")


@pytest.fixture
def mod_comet_txt(basic_comet_df, tmp_path):
    """A tide-like txt file of target psms"""
    out_file = tmp_path / "mod_comet.txt"
    basic_comet_df.to_csv(out_file, sep="\t", index=False)
    return out_file



"""
These tests verify that our q-value calculations are correct.
"""
import logging

import pytest
import numpy as np

from crema.qvalues import tdc, mixmax


# TDC -------------------------------------------------------------------------
@pytest.fixture
def desc_scores():
    """Create a series of descending scores and their q-values"""
    scores = np.array([10, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 2, 1, 1, 1, 1])
    target = np.array([1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0])
    qvals = np.array(
        [
            1 / 4,
            1 / 4,
            1 / 4,
            1 / 4,
            2 / 6,
            2 / 6,
            2 / 6,
            3 / 7,
            3 / 7,
            4 / 7,
            5 / 8,
            5 / 8,
            1,
            1,
            1,
            1,
        ]
    )
    return scores, target, qvals


def test_tdc_descending(desc_scores):
    """Test that q-values are correct for descending scores"""
    scores, target, true_qvals = desc_scores
    dtypes = [np.float64, np.uint8, np.int8, np.float32]
    for dtype in dtypes:
        qvals = tdc(scores.astype(dtype), target, desc=True)
        np.testing.assert_array_equal(qvals, true_qvals)

        qvals = tdc(scores, target.astype(dtype), desc=True)
        np.testing.assert_array_equal(qvals, true_qvals)


def test_tdc_ascending(desc_scores):
    """Test that q-values are correct for ascending scores"""
    scores, target, true_qvals = desc_scores
    scores = -scores
    dtypes = [np.float64, np.uint8, np.int8, np.float32]
    for dtype in dtypes:
        qvals = tdc(scores.astype(dtype), target, desc=False)
        np.testing.assert_array_equal(qvals, true_qvals)

        qvals = tdc(scores, target.astype(dtype), desc=False)
        np.testing.assert_array_equal(qvals, true_qvals)


def test_tdc_non_bool():
    """If targets is not boolean, should get a value error"""
    scores = np.array([1, 2, 3, 4, 5])
    targets = np.array(["1", "0", "1", "0", "blarg"])
    with pytest.raises(ValueError):
        tdc(scores, targets)


def test_tdc_diff_len():
    """If the arrays are different lengths, should get a ValueError"""
    scores = np.array([1, 2, 3, 4, 5])
    targets = np.array([True] * 3 + [False] * 3)
    with pytest.raises(ValueError):
        tdc(scores, targets)


# MixMax -------------------------------------------------------------------------
@pytest.fixture
def mixmax_scores():
    """Increasing score/target arrays with enough depth for mixmax to estimate _q_-values"""
    N = 32
    tgt = 10 + 2 * np.random.randn(N)
    dec = 7 + 2 * np.random.randn(N)

    return (
        np.concatenate([tgt, dec]),
        np.array([True] * len(tgt) + [False] * len(dec)),
    )


@pytest.fixture
def desc_scores_mixmax():
    """Create a series of descending scores and their q-values"""
    scores = np.sort([10, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 2, 1, 1, 1, 1])
    target = np.array([1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0])
    qvals = np.array(
        [
            0,
            0,
            0,
            0,
            np.nan,
            0.0833333,
            0.0833333,
            np.nan,
            0.1122449,
            np.nan,
            0.14671053,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
        ]
    )
    return scores, target, qvals


def do_mixmax(scores, target, desc):
    """Compute the necessary inputs for `mixmax()`"""
    # qvalues.py::calculate_mixmax_qval expects target scores
    # and decoy scores to be sorted from worst to best
    # qvalues.py::mixmax expected combined_sorted scores
    # to be sorted from best to worst
    tgt = np.array(sorted(scores[target.astype(bool)], reverse=desc))
    dec = np.array(sorted(scores[~target.astype(bool)], reverse=desc))
    all_scores = sorted(zip(scores, target), reverse=not desc)

    # Target/decoy score arrays must always be increasing, so invert if necessary
    if desc:
        tgt = -1 * tgt
        dec = -1 * dec

    res = mixmax(
        tgt,
        dec,
        np.array([s for s, _ in all_scores]),
        np.array([t for _, t in all_scores]),
        # **kwargs,
    )

    # Q-values are returned sorted worst to best; reverse them to match best-to-worst order
    # of the combined dataset
    return res[0], res[1][::-1]


def test_mixmax_descending(desc_scores_mixmax):
    """Test that q-values are correct for descending scores"""
    scores, target, true_qvals = desc_scores_mixmax

    tgt_qvals = [q for q, t in zip(true_qvals, target) if t]

    dtypes = [np.float64, np.uint8, np.int8, np.float32]
    for dtype in dtypes:
        qvals = do_mixmax(scores.astype(dtype), target, desc=True)[1]
        assert np.allclose(qvals, tgt_qvals, atol=1e-5)

        qvals = do_mixmax(scores, target.astype(dtype), desc=True)[1]
        assert np.allclose(qvals, tgt_qvals, atol=1e-5)


def test_mixmax_ascending(desc_scores_mixmax):
    """Test that q-values are correct for ascending scores"""
    scores, target, true_qvals = desc_scores_mixmax

    tgt_qvals = [q for q, t in zip(true_qvals, target) if t]

    scores = -scores
    dtypes = [np.float64, np.uint8, np.int8, np.float32]
    for dtype in dtypes:
        qvals = do_mixmax(scores.astype(dtype), target, desc=False)[1]
        assert np.allclose(qvals, tgt_qvals, atol=1e-5)

        qvals = do_mixmax(scores, target.astype(dtype), desc=False)[1]
        assert np.allclose(qvals, tgt_qvals, atol=1e-5)


def test_mixmax_singular(desc_scores):
    """Test that q-values are correct when pi0 == 1.0"""
    scores, target, _ = desc_scores

    dtypes = [np.float64, np.uint8, np.int8, np.float32]
    for dtype in dtypes:
        pi0, qvals = do_mixmax(scores.astype(dtype), target, desc=True)
        assert pi0 == 1.0
        assert all(q == 1.0 for q in qvals)

        pi0, qvals = do_mixmax(scores, target.astype(dtype), desc=True)
        assert pi0 == 1.0
        assert all(q == 1.0 for q in qvals)


def test_mixmax_nonsingular(mixmax_scores):
    """Test that q-values can be computed for pi0 != 1.0"""
    scores, target = mixmax_scores

    dtypes = [np.float64, np.uint8, np.int8, np.float32]
    for dtype in dtypes:
        pi0, qvals = do_mixmax(scores.astype(dtype), target, desc=False)
        assert pi0 != 1.0
        assert any(q != 1.0 for q in qvals)

        pi0, qvals = do_mixmax(scores, target.astype(dtype), desc=False)
        assert pi0 != 1.0
        assert any(q != 1.0 for q in qvals)



"""
This module estimates q-values.
"""
import logging

import numpy as np
import numba as nb

LOGGER = logging.getLogger(__name__)


def tdc(scores, target, desc=True):
    """
    Estimate q-values using target decoy competition.

    Estimates q-values using the simple target decoy competition method.
    For set of target and decoy PSMs meeting a specified score threshold,
    the false discovery rate (FDR) is estimated as:

    ...math:
        FDR = \frac{Decoys + 1}{Targets}

    More formally, let the scores of target and decoy PSMs be indicated as
    :math:`f_1, f_2, ..., f_{m_f}` and :math:`d_1, d_2, ..., d_{m_d}`,
    respectively. For a score threshold :math:`t`, the false discovery
    rate is estimated as:

    ...math:
        E\\{FDR(t)\\} = \frac{|\\{d_i > t; i=1, ..., m_d\\}| + 1}
        {\\{|f_i > t; i=1, ..., m_f|\\}}

    The reported q-value for each PSM is the minimum FDR at which that
    PSM would be accepted.

    Parameters
    ----------
    scores : numpy.ndarray of float
        A 1D array containing the score to rank by

    target : numpy.ndarray of bool
        A 1D array indicating if the entry is from a target or decoy
        hit. This should be boolean, where `True` indicates a target
        and `False` indicates a decoy. `target[i]` is the label for
        `metric[i]`; thus `target` and `metric` should be of
        equal length.

    desc : bool
        Are higher scores better? `True` indicates that they are,
        `False` indicates that they are not.

    Returns
    -------
    numpy.ndarray
        A 1D array with the estimated q-value for each entry. The
        array is the same length as the `scores` and `target` arrays.
    """
    scores = np.array(scores)

    try:
        target = np.array(target, dtype=bool)
    except ValueError:
        raise ValueError("'target' should be boolean.")

    if scores.shape[0] != target.shape[0]:
        raise ValueError("'scores' and 'target' must be the same length")

    # Unsigned integers can cause weird things to happen.
    # Convert all scores to floats to for safety.
    if np.issubdtype(scores.dtype, np.integer):
        scores = scores.astype(np.float_)

    # Sort and estimate FDR
    if desc:
        srt_idx = np.argsort(-scores)
    else:
        srt_idx = np.argsort(scores)

    scores = scores[srt_idx]
    target = target[srt_idx]
    cum_targets = target.cumsum()
    cum_decoys = (~target).cumsum()

    # Handles zeros in denominator
    fdr = np.divide(
        (cum_decoys + 1),
        cum_targets,
        out=np.ones_like(cum_targets, dtype=float),
        where=(cum_targets != 0),
    )

    # Some arrays need to be flipped so that we can loop through from
    # worse to best score.
    fdr = np.flip(fdr)
    scores = np.flip(scores)
    qvals = _fdr2qvalue(scores, fdr)
    qvals = np.flip(qvals)
    qvals = qvals[np.argsort(srt_idx)]

    return qvals


@nb.njit
def _fdr2qvalue(scores, fdr):
    """Quickly calculate q-values.

    Note that the if block logic in this function is to account for multiple
    tied scores. This function assumes that scores and fdr are sorted such
    that they are ordered from worst to best score. Additionally, for a series
    of tied scores, we assume that the first entry in these sorted arrays
    accounts for all of the PSMs that were tied.

    Parameters
    ----------
    scores : np.array
        The scores of the PSMs, sorted from worst to best.
    fdr : np.array
        The calculated FDR of the PSMs, sorted to match scores. Note that
        this is not the same thing as sorting the fdr array itself!

    Returns
    -------
    np.ndarray
        An array of q-values, in the same order as the input scores.
    """
    min_q = 1
    qvals = np.ones(len(fdr))
    start = 0
    for idx in range(len(qvals)):
        if idx < len(qvals) - 1 and scores[idx + 1] == scores[start]:
            continue

        if fdr[start] < min_q:
            min_q = fdr[start]

        qvals[start : idx + 1] = min_q
        start = idx + 1

    return qvals


def mixmax(target_scores, decoy_scores, combined_score, combined_score_target):
    """
    Estimate q-values using mix-max competition.

    TODO
    Estimates q-values using...

    Parameters
    ----------
    target_scores : numpy.ndarray of float
       An array of the best target PSM score per spectrum.
    decoy_scores : numpy.ndarray of float
       An array of the best decoy PSM score per spectrum.
    combined_score : pandas.DataFrame
       A :py:class:`pandas.DataFram` of PSMs. Dataframe includes
       best ranked target and decoy per spectrum.
    combined_score_target: numpy.ndarray of float
       The target/deoy column from the combined_score dataframe

    Returns
    ----------
    pi0 : float
        Estimated pi_zero.
    fdrmod : numpy.ndarray
        Array of calculated q-values.
    """
    # TODO try except and some error checking

    num_targets = target_scores.shape[0]
    num_decoys = decoy_scores.shape[0]

    # calculate p-values from scores
    n_decoys = 1
    pos_same = 0
    neg_same = 0
    pval_list = []
    cur_score = None
    for score, target in zip(combined_score, combined_score_target):
        if target:
            pos_same += 1
        else:
            neg_same += 1

        # TODO Unclear if need if statement. This appears in Percolator code
        # however, adding it makes Crema and Percolator pValList different length
        # if curScore != score:
        for ix in range(0, pos_same):
            pval_list.append(n_decoys + (neg_same * (ix + 1)) / (pos_same + 1))

        n_decoys += neg_same
        neg_same = 0
        pos_same = 0
        cur_score = score
    pval_list = np.array(pval_list) / n_decoys

    # calculate pi0
    pi0 = estimate_pi0(pval_list)

    if pi0 == 1.0:
        # All targets are assumed to be incorrect! Algorithm 1, line 5-6
        LOGGER.debug("FALLBACK: pi0==1.0; all q-values will be 1.0")
        fdrmod = np.full(num_targets, 1.0)  # all q-values are 1
    elif pi0 < 0 or pi0 >= 1 or not np.isfinite(pi0):
        raise ValueError(
            f"Invalid pi0 estimate ({pi0}); unable to proceed FDR estimation!"
        )
    else:
        fdrmod = calculate_mixmax_qval(
            np.array(target_scores), np.array(decoy_scores), pi0
        )

    return (pi0, fdrmod)


@nb.njit
def estimate_pi0(pval_list):
    """
    Estimates pi0. Add description. TODO

    Parameters
    -------
    pval_list : numpy.ndarray of float
        A list of p-values sorted from smallest to largest.

    Returns
    -------
    pi0 : float
        Estimated pi_zero.
    """
    num_lambda = 100
    max_lambda = 0.5
    num_boot = 100

    # LOGGER.debug("pval_list=%s", pval_list)

    n_pval = pval_list.size
    lambda_list = []
    pi0s_list = []
    for idx in range(0, num_lambda):
        cur_lambda = ((idx + 1) / num_lambda) * max_lambda

        # Find the index of the first element in pval_list
        # that is > lambda.
        start = np.searchsorted(pval_list, cur_lambda)
        W1 = n_pval - start
        pi0 = W1 / n_pval / (1.0 - cur_lambda)

        if pi0 > 0.0:
            lambda_list.append(cur_lambda)
            pi0s_list.append(pi0)

    # LOGGER.debug("%s", lambda_list)
    # LOGGER.debug("%s", pi0s_list)

    assert len(pi0s_list) != 0, (
        "Error in the input data: "
        "too good separation between target and decoy PSMs."
    )

    minPi0 = min(pi0s_list)

    mse_list = np.zeros(len(pi0s_list))
    max_size = 1000
    # Examine which lambda level that is most stable under bootstrap
    for i in range(0, num_boot):
        # Create an array of bootstrapped p-values, and sort in ascending order.
        num_draw = min(n_pval, max_size)
        pBoot_list = np.random.choice(pval_list, size=num_draw, replace=True)
        pBoot_list.sort()

        for idx in range(0, len(lambda_list)):
            start = np.searchsorted(pBoot_list, lambda_list[idx])
            W1 = num_draw - start
            pi0Boot = W1 / num_draw / (1.0 - lambda_list[idx])

            # Estimated mean-squared error
            mse_list[idx] += (pi0Boot - minPi0) * (pi0Boot - minPi0)

    # Which index did the iterator go?
    minIdx = np.argmin(mse_list)

    # LOGGER.debug(f"Estimated pi0=%f at lambda=%f (MSE=%f)", pi0s_list[minIdx], lambda_list[minIdx], mse_list[minIdx])

    pi0 = max(min(pi0s_list[minIdx], 1.0), 0.0)
    return pi0


@nb.njit
def calculate_mixmax_qval(target_scores, decoy_scores, pi0):
    """
    Estimate q-values using mix-max.
    """
    # Note that the notation in this function follows the notation found in
    # Percolator, which itself follows the notation found in Supplementary Note
    # 3 (Keich et al., JPR. 2015.). This supplment can be found at
    # http://dx.doi.org/10.1021/acs.jproteome.5b00081.

    # assert pi0 >= 0 and pi0 < 1

    num_targets = target_scores.shape[0]
    num_decoys = decoy_scores.shape[0]

    h_w_le_z = np.zeros(num_decoys)  # histogram for N_{w<=z}
    h_z_le_z = np.zeros(num_decoys)  # histogram for N_{z<=z}

    for j in range(0, num_decoys):
        h_w_le_z[j] = np.searchsorted(
            target_scores, decoy_scores[j], side="right"
        )
        h_z_le_z[j] = np.searchsorted(
            decoy_scores, decoy_scores[j], side="right"
        )

    fdrmod = np.zeros(num_targets)
    estPx_lt_zj = 0.0
    E_f1_mod_run_tot = 0.0
    j = num_decoys - 1
    n_z_ge_w = 0
    n_w_ge_w = 0
    prev_fdr = -1

    for i in range(num_targets - 1, -1, -1):
        while j >= 0 and decoy_scores[j] >= target_scores[i]:
            cnt_w = h_w_le_z[j]
            cnt_z = h_z_le_z[j]

            estPx_lt_zj = (cnt_w - pi0 * cnt_z) / ((1.0 - pi0) * cnt_z)
            if estPx_lt_zj > 1:
                estPx_lt_zj = 1.0
            elif estPx_lt_zj < 0:
                estPx_lt_zj = 0.0

            E_f1_mod_run_tot += estPx_lt_zj * (1.0 - pi0)
            n_z_ge_w += 1
            j -= 1

        # not sure if this faster or if original while loop is faster
        # AssignConfidence.cpp:1225
        n_w_ge_w = (target_scores >= target_scores[i]).sum()
        fdr = (n_z_ge_w * pi0 + E_f1_mod_run_tot) / n_w_ge_w

        if fdr > 1:
            fdr = 1.0
        fdrmod[i] = fdr

    # convert qvalues to fdr
    return _fdr2qvalue(target_scores, fdrmod)



"""A parser for the Tide tab-delimited format"""
import re
import logging

import pandas as pd

from .txt import read_txt
from .. import utils

LOGGER = logging.getLogger(__name__)


def read_tide(
    txt_files, pairing_file_name=None, decoy_prefix="decoy_", copy_data=True
):
    """Read peptide-spectrum matches (PSMs) from Tide tab-delimited files.

    Parameters
    ----------
    txt_files : str, pandas.DataFrame or tuple of str
        One or more collection of PSMs in the Tide tab-delimited format.
    pairing_file_name : str, optional
        A tab-delimited file that explicity pairs target and decoy peptide
        sequences. Requires one column labled 'target' that contains target
        sequences and a second colun labeled 'decoy' that contains decoy
        sequences. This file can be generated by setting --peptide-list=T
        in tide-index.
    decoy_prefix : str, optional
        The prefix used to indicate a decoy protein in the protein column.
        Default value is 'decoy_'.
    copy_data : bool, optional
        If true, a deep copy of the data is created. This uses more memory, but
        is safer because it prevents accidental modification of the underlying
        data. This argument only has an effect when `txt_files` is a
        :py:class:`pandas.DataFrame`

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object containing the parsed
        PSMs.
    """
    target = "target/decoy"
    peptide = "sequence"
    spectrum = ["file", "scan"]
    pairing = "original target sequence"
    protein = "protein id"
    protein_delim = ","

    # Possible score columns output by Tide.
    scores = {
        "sp score",
        "delta_cn",
        "delta_lcn",
        "xcorr score",
        "exact p-value",
        "refactored xcorr",
        "res-ev p-value",
        "combined p-value",
        "tailor score",
    }
    scores_all = scores

    # Keep only Tide scores that exist in all of the files.
    if isinstance(txt_files, pd.DataFrame):
        scores = scores.intersection(set(txt_files.columns))
    else:
        txt_files = utils.listify(txt_files)
        for txt_file in txt_files:
            with open(txt_file) as txt_ref:
                cols = txt_ref.readline().rstrip().split("\t")
                scores = scores.intersection(set(cols))

    if not scores:
        raise ValueError(
            "Could not find any of the Tide score columns in all of the files."
            f" The columns Crema looks for are {', '.join(list(scores_all))}"
        )

    scores = list(scores)

    # Read in the files:
    fields = [*spectrum, peptide, target, *scores, pairing, protein]
    if isinstance(txt_files, pd.DataFrame):
        data = txt_files.copy(deep=copy_data).loc[:, fields]
    else:
        data = pd.concat(
            [utils.parse_psms_txt(f, fields, False) for f in txt_files]
        )

    psms = read_txt(
        data,
        target_column=target,
        spectrum_columns=spectrum,
        score_columns=scores,
        peptide_column=peptide,
        protein_column=protein,
        protein_delim=protein_delim,
        sep="\t",
        copy_data=False,
    )

    # always pair target and decoys for Tide
    if pairing_file_name == None:  # implicit pairing
        psms._peptide_pairing = _create_pairing(data)
    else:  # explicit pairing
        psms._peptide_pairing = utils.create_pairing_from_file(
            pairing_file_name
        )

    # Remove the start position of peptide in protein if present
    # This looks like "protName(XX)"
    # Remove decoy prefix from protein ID
    protein_column = psms.proteins
    new_protein_column = protein_column.str.replace(
        "\\([^()]*\\)", "", regex=True
    )
    new_protein_column = new_protein_column.str.replace(
        decoy_prefix, "", regex=True
    )
    psms.set_protein_column(new_protein_column)

    return psms


def _create_pairing(pairing_data):
    """Parse a single Tide dataframe to implicity pair target and
    decoy sequences.

    Parameters
    ----------
    pairing_data : pandas.DataFrame
        A collection of PSMs with the necessary columns to create a
        target/decoy peptide pairing. Required columns are "peptide mass",
        "sequence", "target/decoy", "original target sequence"

    Returns
    -------
    pairing : dict
        A map of target and decoy peptide sequence pairings. Targets with
        missing decoys will not be included among the keys.

    """
    # ensure pairing_data dataframe contains all necessary columns
    seq = "original target sequence"
    req_fields = [
        "sequence",
        "target/decoy",
        "original target sequence",
    ]

    if not set(req_fields).issubset(pairing_data.columns):
        miss = ", ".join(set(req_fields) - set(pairing_data.columns))
        raise ValueError(
            f"Required columns for peptide pairing were not detected: {miss}"
        )

    pairing_data = pairing_data.loc[:, req_fields]
    pairing_data = (
        pairing_data.sample(frac=1)
        .drop_duplicates(["sequence"])
        .reset_index(drop=False)
    )

    # Add a column of the sorted peptide:
    pairing_data["mods"] = (
        pairing_data["sequence"]
        .str.split("(?=[A-Z])")
        .apply(lambda x: "".join(sorted(x)))
    )

    # Split targets and decoys:
    is_decoy = pairing_data["target/decoy"] == "decoy"
    pairing_data = pairing_data.drop("target/decoy", axis=1)
    targets = pairing_data.loc[~is_decoy, :].copy()
    decoys = pairing_data.loc[is_decoy, :].copy()

    # Strip the target sequence modifications:
    targets[seq] = targets["sequence"].str.replace(r"\[.*?\]", "", regex=True)

    # Add an 'ord' column to disambiguate multiple matches per peptide:
    targets["ord"] = targets.groupby([seq, "mods"])["sequence"].rank("first")
    decoys["ord"] = decoys.groupby([seq, "mods"])["sequence"].rank("first")

    # Inner join the DataFrames to induce a pairing.
    # Targets with a missing decoy will be dropped.
    # Decoys with a missing target will be dropped.
    merged = pd.merge(
        targets,
        decoys,
        how="inner",
        on=[seq, "mods", "ord"],
        suffixes=["_t", "_d"],
    )

    return merged.set_index("sequence_t").loc[:, "sequence_d"].to_dict()



"""These tests verify that the crema CLI works as expected."""
from pathlib import Path
import subprocess


def test_cli_basic(basic_tide_txt, tmp_path):
    """Test that the basic cli works."""
    cmd = ["crema", "--output_dir", tmp_path, "-e", "0.5", basic_tide_txt]
    subprocess.run(cmd, check=True)
    assert Path(tmp_path, "crema.psms.txt").exists()
    assert Path(tmp_path, "crema.peptides.txt").exists()
    assert Path(tmp_path, "crema.proteins.txt").exists()
    assert Path(tmp_path, "crema.log.txt").exists()


def test_cli_custom_root(basic_tide_txt, tmp_path):
    """Test that the cli works with custom file root."""
    cmd = ["crema", "-o", tmp_path, "-f", "myFileRoot", "-e", "0.5"]
    cmd.append(basic_tide_txt)
    subprocess.run(cmd, check=True)
    assert Path(tmp_path, "myFileRoot.crema.psms.txt").exists()
    assert Path(tmp_path, "myFileRoot.crema.peptides.txt").exists()
    assert Path(tmp_path, "myFileRoot.crema.proteins.txt").exists()
    assert Path(tmp_path, "myFileRoot.crema.log.txt").exists()


def test_real(real_tide_txt, tmp_path):
    """Test that crema works on real Crux results"""
    cmd = ["crema", "--output_dir", tmp_path] + list(real_tide_txt)
    subprocess.run(cmd, check=True)
    assert Path(tmp_path, "crema.psms.txt").exists()
    assert Path(tmp_path, "crema.peptides.txt").exists()
    assert Path(tmp_path, "crema.proteins.txt").exists()
    assert Path(tmp_path, "crema.log.txt").exists()



"""
These are unit tests for functions within parsers.py:
"""
import pytest
import pandas as pd

import crema


def test_read_tide(real_tide_txt):
    """Test that we parse Tide files correctly"""
    psms = crema.read_tide(real_tide_txt)
    assert isinstance(psms.data, pd.DataFrame)
    assert psms.data.shape == (21818, 12)
    assert list(psms.spectra.columns) == ["file", "scan"]

    scores = {
        "refactored xcorr",
        "combined p-value",
        "exact p-value",
        "sp score",
        "delta_cn",
        "delta_lcn",
        "res-ev p-value",
    }
    assert set(psms.score_columns) == scores
    assert psms.scores.shape == (21818, len(scores))
    assert psms.targets.shape == (21818,)
    assert psms.targets.sum() == 10909
    assert (~psms.targets).sum() == 10909


def test_read_tide_peptide_pairing(mod_target_tide_txt, mod_decoy_tide_txt):
    """Test that peptide pairing is correctly created when parsing Tide files"""
    expected_peptide_pairing = {
        "ALLSLR": "ALLLSR",
        "QTPPAR": "QTAPPR",
        "GEVPN[0.98]R": "GN[0.98]PEVR",
        "GGHMDR": "GDMGHR",
    }
    psms = crema.read_tide(
        [mod_decoy_tide_txt, mod_target_tide_txt],
    )
    assert expected_peptide_pairing == psms.peptide_pairing


def test_read_comet_peptide_pariring(mod_comet_txt):
    """Test that peptide paiing is correctly create when parsing comet file"""
    expected_peptide_pairing = {
        "A[15.9949]PPLE": "A[15.9949]LPPE",
        "BANANA": "BNANAA",
        "CH[15.9949]E[15.9949]RRY": "CRRE[15.9949]H[15.9949]Y",
        "DURIAN": "DAIRUN",
        "EGGPLANT": "ENALPGGT",
        "FIGS": "FGIS",
        "GRAPE": "GPARE",
        "H[15.9949]ONEYDE[15.9949]W": "H[15.9949]E[15.9949]DYENOW",
        "IC[15.9949]ES": "IEC[15.9949]S",
        "JAMS": "JMAS",
    }
    psms = crema.read_comet(mod_comet_txt)
    for key in expected_peptide_pairing:
        print(key, expected_peptide_pairing[key], psms.peptide_pairing[key])
    assert expected_peptide_pairing == psms.peptide_pairing


def test_read_txt(basic_tide_csv):
    """Test that we can read generic delimited files"""
    psms = crema.read_txt(
        basic_tide_csv,
        target_column="target/decoy",
        spectrum_columns="scan",
        score_columns=["combined p-value", "x"],
        peptide_column="sequence",
        protein_column="protein id",
        protein_delim=",",
        sep=",",
    )
    assert isinstance(psms.data, pd.DataFrame)
    assert psms.data.shape == (10, 6)
    assert psms.spectra.columns == ["scan"]
    assert set(psms.score_columns) == {"combined p-value", "x"}
    assert psms.peptides.name == "sequence"
    assert psms.proteins.name == "protein id"
    assert psms.targets.shape == (10,)
    assert psms.targets.sum() == 6
    assert (~psms.targets).sum() == 4


def test_read_msgf(basic_msgf_txt):
    psms = crema.read_msgf(basic_msgf_txt)
    assert isinstance(psms.data, pd.DataFrame)
    assert psms.data.shape == (10, 9)
    assert list(psms.spectra.columns) == ["#SpecFile", "ScanNum"]

    scores = {
        "DeNovoScore",
        "MSGFScore",
        "SpecEValue",
        "EValue",
    }
    assert set(psms.score_columns) == scores
    assert psms.scores.shape == (10, len(scores))
    assert psms.targets.shape == (10,)
    assert psms.targets.sum() == 7
    assert (~psms.targets).sum() == 3


def test_read_msamanda(basic_msamanda_txt):
    psms = crema.read_msamanda(basic_msamanda_txt)
    assert psms.data.shape == (10, 7)
    assert list(psms.spectra.columns) == ["Filename", "Scan Number"]

    scores = {"Amanda Score", "Weighted Probability"}
    assert set(psms.score_columns) == scores
    assert psms.scores.shape == (10, len(scores))
    assert psms.targets.shape == (10,)
    assert psms.targets.sum() == 4
    assert (~psms.targets).sum() == 6


def test_read_msfragger_pepxml(real_msfragger_pepxml):
    psms = crema.read_msfragger(real_msfragger_pepxml)
    assert psms.data.shape == (99, 8)
    assert list(psms.spectra.columns) == ["ms_data_file", "scan"]

    scores = {"nextscore", "expect", "hyperscore"}
    assert set(psms.score_columns) == scores
    assert psms.scores.shape == (99, len(scores))
    assert psms.targets.shape == (99,)
    assert psms.targets.sum() == 94
    assert (~psms.targets).sum() == 5


def test_read_mztab(real_mztab):
    # This test suposed to raise error as mzID file does not
    # contain decoy PSMs
    with pytest.raises(ValueError):
        psms = crema.read_mztab(real_mztab)


def test_read_pepxml(real_pepxml):
    try:
        psms = crema.read_pepxml(real_pepxml, "decoy_")
    except Exception as exc:
        assert False, f"'test_read_pepxml' raised an exception {exc}"



"""
These are unit tests for the PSM Dataset Class:
"""
import pytest
import numpy as np
import pandas as pd

from crema import PsmDataset


@pytest.fixture
def simple_df(basic_tide_df):
    """A simple dataframe of PSMs"""
    df = basic_tide_df
    df["target"] = df["target/decoy"].replace({"target": True, "decoy": False})
    return df.drop(columns="target/decoy")


def test_create_object(simple_df):
    """Ensures that a PsmDataset object can be initialized properly."""
    psms = PsmDataset(
        psms=simple_df,
        target_column="target",
        spectrum_columns=["file", "scan"],
        score_columns=["combined p-value", "x"],
        peptide_column="sequence",
        protein_column="protein id",
        protein_delim=",",
    )
    assert isinstance(psms, PsmDataset)


def test_properties(simple_df):
    """Check the public properties of the PsmDataset object."""
    psms = PsmDataset(
        psms=simple_df,
        target_column="target",
        spectrum_columns=["file", "scan"],
        score_columns=["combined p-value", "x"],
        peptide_column="sequence",
        protein_column="protein id",
        protein_delim=",",
    )

    # TODO remove after Will code review
    # This is not true anymore because columns are used as input that
    # are not found in output (namely "original target sequence).
    # pd.testing.assert_frame_equal(psms.data, simple_df, check_like=True)
    assert list(psms.spectra.columns) == ["file", "scan"]
    assert psms.score_columns == ["combined p-value", "x"]
    assert psms.peptides.name == "sequence"
    assert psms.proteins.name == "protein id"
    assert psms._protein_delim == ","
    pd.testing.assert_frame_equal(
        psms.scores, simple_df.loc[:, ["combined p-value", "x"]]
    )
    np.testing.assert_array_equal(
        psms.targets, simple_df.loc[:, "target"].values
    )


def test_getitem(simple_df):
    """Test the __getitem__ method"""
    psms = PsmDataset(
        psms=simple_df,
        target_column="target",
        spectrum_columns=["file", "scan"],
        score_columns=["combined p-value", "x"],
        peptide_column="sequence",
        protein_column="protein id",
        protein_delim=",",
    )

    pd.testing.assert_series_equal(psms["scan"], simple_df["scan"])
    pd.testing.assert_series_equal(
        psms["combined p-value"],
        simple_df["combined p-value"],
    )


def test_find_best_score(simple_df):
    """Test the find_best_score() method"""
    psms = PsmDataset(
        psms=simple_df,
        target_column="target",
        spectrum_columns=["scan", "spectrum precursor m/z"],
        score_columns=["combined p-value", "x"],
        peptide_column="sequence",
        protein_column="protein id",
        protein_delim=",",
    )

    score, npass, desc = psms.find_best_score(eval_fdr=0.4)
    assert score == "x"
    assert npass == 4
    assert desc



"""Writer to save results in a tab-delmited format"""
from pathlib import Path
from collections import defaultdict

import pandas as pd


def to_txt(
    conf, output_dir=None, file_root=None, sep="\t", decoys=False, precision=6
):
    """Save confidence estimates to delimited text files.

    Write the confidence estimates for each of the available levels
    (i.e. PSMs, peptides, proteins) to separate flat text files using the
    specified delimiter. If more than one collection of confidence estimates
    is provided, they will be combined, yielding a single file for each level
    specified by either dataset.

    Parameters
    ----------
    conf : Confidence object or tuple of Confidence objects
        One or more :py:class:`~crema.confidence.Confidence` objects.
    output_dir : str or None, optional
        The directory in which to save the files. :code:`None` will use the
        current working directory.
    file_root : str or None, optional
        An optional prefix for the confidence estimate files. The suffix will
        always be "crema.{level}.txt" where "{level}" indicates the level at
        which confidence estimation was performed (i.e. PSMs, peptides,
        proteins).
    sep : str, optional
        The delimiter to use.
    decoys : bool, optional
        Save decoys confidence estimates as well?
    precision : int, optional
        Precision for float values.

    Returns
    -------
    list of str
        The paths to the saved files.

    """
    try:
        assert not isinstance(conf, str)
        iter(conf)
    except TypeError:
        conf = [conf]
    except AssertionError:
        raise ValueError("'conf' should be a Confidence object, not a string.")

    file_base = "crema"
    if file_root is not None:
        file_base = file_root + "." + file_base
    if output_dir is not None:
        file_base = Path(output_dir, file_base)

    results = defaultdict(list)
    for res in conf:
        for level, qval_list in _get_level_data(res, decoys).items():
            results[level] += qval_list

    out_files = []
    for level, qval_list in results.items():
        out_file = str(file_base) + f".{level}.txt"
        pd.concat(qval_list).to_csv(
            out_file, sep=sep, index=False, float_format=f"%.{precision}f"
        )
        out_files.append(out_file)

    return out_files


def _get_level_data(conf, decoys):
    """Return the dataframes for each level.

    Parameters
    ----------
    conf : a Confidence object
        A LinearConfidence object.
    decoys : bool
        Should decoys be included?

    Returns
    -------
    Dict
        Each entry contains a level, dataframe pair.
    """
    results = defaultdict(list)
    for level, qvals in conf.confidence_estimates.items():
        if qvals is None:
            continue

        results[level].append(qvals)

    if decoys:
        for level, qvals in conf.decoy_confidence_estimates.items():
            if qvals is None:
                continue

            results[f"decoy.{level}"].append(qvals)

    return results



"""Utility functions that are used in multiple modules"""
import pandas as pd
import logging

LOGGER = logging.getLogger(__name__)


def listify(obj):
    """Create list containing an object if it is not already a list."""
    try:
        assert not isinstance(obj, str)
        iter(obj)
    except (AssertionError, TypeError):
        obj = [obj]

    return list(obj)


def new_column(name, df):
    """Add a new column, ensuring a unique name"""
    new_name = name
    cols = set(df.columns)
    i = 0
    while new_name in cols:
        new_name = name + "_" + str(i)
        i += 1

    return new_name


def create_pairing_from_file(pairing_file_name):
    """Parse a single file that explicity pairs target and decoy sequences.

    Parameters
    ----------
    pairing_file_name : str
        A tab-delimited file that explicity pairs target and decoy peptide
        sequences. Requires one column labled 'target' that contains target
        sequences and a second column labeled 'decoy(s)' that contains decoy
        sequences. For analying fles generated by Tide, this file can be
        generated by setting --peptide-list=T in tide-index.

    Returns
    -------
    pairing : dict
        A map of target and decoy peptide sequence pairings. Targets with
        missing decoys will not be included among the keys.

    """
    pairing_file = pd.read_csv(pairing_file_name, sep="\t")

    # ensure pairing_file dataframe contains all necessary columns
    target_field = "target"
    decoy_field = "decoy(s)"
    req_fields = [
        target_field,
        decoy_field,
    ]

    if not set(req_fields).issubset(pairing_file.columns):
        miss = ", ".join(set(req_fields) - set(pairing_file.columns))
        raise ValueError(
            f"Required columns for peptide pairing were not detected: {miss}"
        )

    # drop targets that do not have corresponding decoys
    pairing_file = pairing_file[pairing_file[decoy_field] != ""]
    return dict(zip(pairing_file[target_field], pairing_file[decoy_field]))


def parse_psms_txt(txt_file, cols, skip_line):
    """Parse a single tab-delimited file

    Parameters
    ----------
    txt_file : str
        The tab-delimited file of PSMs to read.
    cols : list of str
        The columns to parse.
    skip_line : bool
        If true, skip reading the first line.

    Returns
    -------
    pandas.DataFrame
        A :py:class:`pandas.DataFrame` containing the parsed PSMs
    """
    LOGGER.info("Reading PSMs from %s...", txt_file)

    # Because skip_line is a boolean:
    return pd.read_csv(
        txt_file,
        sep="\t",
        skiprows=int(skip_line),
        usecols=lambda c: c in cols,
    )



"""
This module contains the parser for PSMs in pepXML format.
This code is heavily based on Will Fondrie's Mokapot pepxml parser code
"""
import logging
from lxml import etree
from functools import partial

import pandas as pd
import itertools
import re

from ..utils import listify
from ..dataset import PsmDataset

LOGGER = logging.getLogger(__name__)


def read_pepxml(pepxml_files, decoy_prefix):
    """Read peptide-spectrum matches (PSMs) from pepXML files.

    Parameters
    ----------
    pepxml_files : str or tuple of str
       One or more collections of PSMs in the pepXML format.
    decoy_prefix : str
       The prefix used to indicate a decoy protein in the
       description lines of the FASTA file.

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSMs from the pepxml file.
    """
    pepxml_files = listify(pepxml_files)

    # Create a dataframe from the PSMs in the pepXML files.
    psms = pd.concat([_parse_pepxml(f, decoy_prefix) for f in pepxml_files])

    # Initialize column names from pepXML standard specifications
    spectrum_col = ["ms_data_file", "scan"]
    score_col = [c for c in psms.columns if "search_engine_score" in c]
    target_col = "label"
    sequence_col = "peptide"
    protein_col = "proteins"
    protein_delim = ","

    # Check that all column headers are valid, otherwise, throw error
    if len(set(spectrum_col) & set(psms.columns)) < len(spectrum_col):
        raise KeyError(
            "The pepXML file does not contain the columns that define a "
            f"spectrum. These are: {', '.join(spectrum_col)}."
        )

    if sequence_col not in psms.columns:
        raise KeyError(
            "The pepXML file does not contain the columns to specify "
            "peptide sequence."
        )

    if target_col not in psms.columns:
        raise KeyError(
            "The pepXML file does not contain the column that specifies "
            f"whether a PSM is a target or decoy, {target_col}"
        )

    if not score_col:
        raise ValueError(
            "No columns containing search engine scores were detected. These "
            "start with 'search_engine_score*'."
        )

    if not protein_col:
        raise ValueError(
            "The pepXML file does not contain the columns to specify "
            "protein sequence."
        )

    # Remove "search_engine_score:" from column name and score_col
    # and convert scores to float
    psms.columns = psms.columns.str.replace("search_engine_score:", "")
    score_col = [re.sub("search_engine_score:", "", c) for c in score_col]
    psms[score_col] = psms[score_col].astype(float)

    return PsmDataset(
        psms=psms,
        target_column=target_col,
        spectrum_columns=spectrum_col,
        score_columns=score_col,
        peptide_column=sequence_col,
        protein_column=protein_col,
        protein_delim=protein_delim,
        copy_data=False,
    )


def _parse_pepxml(pepxml_file, decoy_prefix):
    """Parse a single pepXML file using lxml into a DataFrame

    Parameters
    ----------
    pepxml_file : str
        The pepXML file to read.
    decoy_prefix : str
        The prefix used to indicate a decoy protein in the
        description lines of the FASTA file.

    Returns
    -------
    pandas.DataFrame
        A :py:class:`pandas.DataFrame` containing the parsed PSMs.
    """
    LOGGER.info("Reading PSMs from %s...", pepxml_file)
    parser = etree.iterparse(str(pepxml_file), tag="{*}msms_run_summary")
    parse_fun = partial(_parse_msms_run, decoy_prefix=decoy_prefix)
    spectra = map(parse_fun, parser)
    try:
        psms = itertools.chain.from_iterable(spectra)
        df = pd.DataFrame.from_records(itertools.chain.from_iterable(psms))
        df["ms_data_file"] = df["ms_data_file"].astype("category")
    except etree.XMLSyntaxError:
        raise ValueError(
            f"{pepxml_file} is not a PepXML file or is malformed."
        )
    return df


def _parse_msms_run(msms_run, decoy_prefix):
    """Parse a single MS/MS run.

    Each of these corresponds to a raw MS data file.

    Parameters
    ----------
    msms_run: tuple of anything, lxml.etree.Element
        The second element of the tuple should be the XML element for a single
        msms_run. The first is not used, but is necessary for compatibility
        with using :code:`map()`.
    decoy_prefix : str
        The prefix used to indicate a decoy protein in the description lines of
        the FASTA file.
    Yields
    ------
    dict
        A dictionary describing all of the PSMs in a run.
    """
    msms_run = msms_run[1]
    ms_data_file = msms_run.get("base_name")
    run_ext = msms_run.get("raw_data")
    if not ms_data_file.endswith(run_ext):
        ms_data_file += run_ext

    run_info = {"ms_data_file": ms_data_file}
    for spectrum in msms_run.iter("{*}spectrum_query"):
        yield _parse_spectrum(spectrum, run_info, decoy_prefix)


def _parse_spectrum(spectrum, run_info, decoy_prefix):
    """Parse the PSMs for a single mass spectrum
    Parameters
    ----------
    spectrum : lxml.etree.Element
        The XML element for a single
    run_info : dict
        The parsed run data.
    decoy_prefix : str
        The prefix used to indicate a decoy protein in the description lines of
        the FASTA file.
    Yields
    ------
    dict
        A dictionary describing all of the PSMs for a spectrum.
    """
    spec_info = run_info.copy()
    spec_info["scan"] = int(spectrum.get("end_scan"))
    for psms in spectrum.iter("{*}search_result"):
        for psm in psms.iter("{*}search_hit"):
            yield _parse_psm(psm, spec_info, decoy_prefix=decoy_prefix)


def _parse_psm(psm_info, spec_info, decoy_prefix):
    """Parse a single PSM
    Parameters
    ----------
    psm_info : lxml.etree.Element
        The XML element containing information about the PSM.
    spec_info : dict
        The parsed spectrum data.
    decoy_prefix : str
        The prefix used to indicate a decoy protein in the description lines of
        the FASTA file.
    Returns
    -------
    dict
        A dictionary containing parsed data about the PSM.
    """
    psm = spec_info.copy()
    psm["peptide"] = psm_info.get("peptide")
    psm["proteins"] = [psm_info.get("protein").split(" ")[0]]
    psm["label"] = not psm["proteins"][0].startswith(decoy_prefix)

    queries = [
        "{*}modification_info",
        "{*}search_score",
        "{*}alternative_protein",
    ]
    for element in psm_info.iter(*queries):
        if "modification_info" in element.tag:
            offset = 0
            mod_pep = psm["peptide"]
            for mod in element.iter("{*}mod_aminoacid_mass"):
                idx = offset + int(mod.get("position"))
                mass = mod.get("mass")
                mod_pep = mod_pep[:idx] + "[" + mass + "]" + mod_pep[idx:]
                offset += 2 + len(mass)

            psm["peptide"] = mod_pep

        elif "alternative_protein" in element.tag:
            psm["proteins"].append(element.get("protein").split(" ")[0])
            if not psm["label"]:
                psm["label"] = not psm["proteins"][-1].startswith(decoy_prefix)

        else:
            psm["search_engine_score:" + element.get("name")] = element.get(
                "value"
            )

    psm["proteins"] = ",".join(psm["proteins"])
    return psm



"""This module contains the parser for PSMs in mzTab format"""
import logging

import pandas as pd
from pyteomics.mztab import MzTab

from ..utils import listify
from ..dataset import PsmDataset

LOGGER = logging.getLogger(__name__)


def read_mztab(mztab_files):
    """Read peptide-spectrum matches (PSMs) from mzTab files.

    Parameters
    ----------
    mztab_files : str or tuple of str
        One or more collections of PSMs in the mzTab format.

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSMs from the mzTab file.
    """
    mztab_files = listify(mztab_files)

    # Create a dataframe from the PSMs in the mzTab files.
    psms = pd.concat([_parse_psms(f) for f in mztab_files])

    # Initialize column names from mzTab standard specifications
    spectrum_col = ["spectra_ref"]
    score_col = [c for c in psms.columns if "search_engine_score" in c]
    target_col = "opt_global_cv_MS:1002217_decoy_peptide"
    sequence_col = "sequence"
    protein_col = "accession"  # TODO check if correct
    delim_col = ";"  # TODO check if correct
    mod_col = "modifications"

    # Check that all column headers are valid, otherwise, throw error
    if len(set(spectrum_col) & set(psms.columns)) < len(spectrum_col):
        raise KeyError(
            "The mzTab file does not contain the columns that define a "
            f"spectrum. These are: {', '.join(spectrum_col)}."
        )

    if sequence_col not in psms.columns or mod_col not in psms.columns:
        raise KeyError(
            "The mzTab file does not contain the columns to specify "
            "peptide sequence and modifications."
        )

    if target_col not in psms.columns:
        raise KeyError(
            "The mzTab file does not contain the column that specifies "
            f"whether a PSM is a target or decoy, {target_col}"
        )

    if not score_col:
        raise ValueError(
            "No columns containing search engine scores were detected. These "
            "start with 'search_engine_score*'."
        )

    # Create the necesssary columns
    psms["peptide"] = psms[sequence_col] + "[" + psms[mod_col] + "]"
    psms["target"] = ~psms[target_col].astype(bool)

    # Keep only the relevant columns
    columns = spectrum_col + score_col + ["peptide", "target"] + [protein_col]
    psms = psms.loc[:, columns]

    return PsmDataset(
        psms=psms,
        target_column="target",
        spectrum_columns=spectrum_col,
        score_columns=score_col,
        peptide_column="peptide",
        protein_column=protein_col,
        protein_delim=delim_col,
        copy_data=False,
    )


def _parse_psms(mztab_file):
    """Parse a single mzTab file using Pyteomics

    Parameters
    ----------
    mztab_file : str
        The mzTab file to read.

    Returns
    -------
    pandas.DataFrame
        A :py:class:`pandas.DataFrame` containing the parsed PSMs.
    """
    LOGGER.info("Reading PSMs from %s...", mztab_file)
    return MzTab(str(mztab_file)).spectrum_match_table



"""Initialize the crema package."""
try:
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version("crema-ms")
    except PackageNotFoundError:
        pass

except ImportError:
    from pkg_resources import get_distribution, DistributionNotFound

    try:
        __version__ = get_distribution("crema-ms").version
    except DistributionNotFound:
        pass

# Here is where we can export public functions and classes.
from .dataset import PsmDataset
from .parsers.tide import read_tide
from .parsers.msgf import read_msgf
from .parsers.msamanda import read_msamanda
from .parsers.msfragger import read_msfragger
from .parsers.comet import read_comet
from .parsers.txt import read_txt
from .parsers.mztab import read_mztab
from .parsers.pepxml import read_pepxml
from .confidence import TdcConfidence, assign_confidence
from .writers.txt import to_txt



"""The :py:class:`Confidence` class is used to define a collection of
peptide-spectrum matches with calculated false discovery rates (FDR) and q-values.
"""
import logging
import numpy as np
import pandas as pd
from abc import ABC, abstractmethod

from . import qvalues
from . import utils
from .writers.txt import to_txt

np.random.seed(0)

LOGGER = logging.getLogger(__name__)


def assign_confidence(
    psms,
    score_column=None,
    desc=None,
    eval_fdr=0.01,
    method="tdc",
    pep_fdr_type="psm-peptide",
    prot_fdr_type="best",
    threshold=0.01,
):
    """Assign confidence estimates to a collection of peptide-spectrum matches.

    Parameters
    ----------
    psms : PsmDataset or list of PsmDataset objects
        The collections of PSMs
    score_column : str, optional
        The score by which to rank the PSMs for confidence estimation. If
        :code:`None`, the score that yields the most PSMs at the specified
        false discovery rate threshold (`eval_fdr`), will be used.
    desc : bool, optional
        True if higher scores better, False if lower scores are better.
        If None, crema will try both and use the
        choice that yields the most PSMs at the specified false discovery
        rate threshold (`eval_fdr`). If `score_column` is :code:`None`,
        this parameter is ignored.
    eval_fdr : float, optional
        The false discovery rate threshold used to evaluate the best
        `score_column` and `desc` to choose. This should range from 0 to 1.
    method : {"tdc"}, optional
        The method for crema to use when calculating the confidence estimates.
    pep_fdr_type : {"psm-only","peptide-only",psm-peptide"}, optional
        The method for Crema to use when calculating peptide level confidence
        estimates.
    prot_fdr_type : {"best", "combine"}, optional
        The method for crema to use when calculating protein level confidence
        estimates. Default is "best".
    threshold : float or "q-value", optional
        The FDR threshold for accepting discoveries. Default is 0.01. If
        "q-value" is chosen, then "accept" column is replaced with
        "crema q-value".

    Returns
    -------
    Confidence object or List of Confidence objects
        The confidence estimates for each PsmDataset.
    """
    if isinstance(psms, str):
        raise ValueError("'psms' should be a PsmDataset object, not a string.")

    # TODO unable to check whether psms is type PsmDataset w/o circular import
    try:
        assert isinstance(psms, list)
    except (AssertionError, TypeError):
        psms = [psms]

    confs = []
    for dset in psms:
        conf = dset.assign_confidence(
            score_column=score_column,
            desc=desc,
            eval_fdr=eval_fdr,
            method=method,
            pep_fdr_type=pep_fdr_type,
            prot_fdr_type=prot_fdr_type,
            threshold=threshold,
        )
        confs.append(conf)

    if len(confs) == 1:
        return confs[0]

    return confs


class Confidence(ABC):
    """Estimate statistical confidence estimates for a collection of PSMs.

    This should be a good base class for most of the methods, but not one
    directly called by a user.

    Parameters
    ----------
    psms : a PsmDataset object
        A collection of PSMs
    score_column : str, optional
        The score by which to rank the PSMs for confidence estimation. If
        :code:`None`, the score that yields the most PSMs at the specified
        false discovery rate threshold (`eval_fdr`), will be used.
    desc : bool, optional
        True if higher scores better, False if lower scores are better.
        If None, crema will try both and use the
        choice that yields the most PSMs at the specified false discovery
        rate threshold (`eval_fdr`). If `score_column` is :code:`None`,
        this parameter is ignored.
    eval_fdr : float, optional
        The false discovery rate threshold used to evaluate the best
        `score_column` and `desc` to choose. This should range from 0 to 1.
    pep_fdr_type : {"psm-only","peptide-only",psm-peptide"}, optional
        The method for Crema to use when calculating peptide level confidence
        estimates.
    threshold : float or "q-value", optional
        The FDR threshold for accepting discoveries. Default is 0.01. If
        "q-value" is chosen, then "accept" column is replaced with
        "crema q-value".

    Attributes
    ----------
    data : pandas.DataFrame
    dataset : crema.PsmDataset
    levels : list of str
    confidence_estimates : Dict
        A dictionary containing the confidence estimates at each level, each
        as a :py:class:`pandas.DataFrame`.
    decoy_confidence_estimates : Dict
        A dictionary containing the confidence estimates for the decoy hits at
        each level, each as a :py:class:`pandas.DataFrame`

    :meta private:

    """

    _level_labs = {
        "psms": "PSMs",
        "peptides": "Peptides",
        "proteins": "Proteins",
    }

    @abstractmethod
    def _assign_confidence(self):
        """How should confidence estimates be assigned?

        Each method will have its own way of doing this. The results
        should be saved in the `confidence_estimates` and
        `decoy_confidence_estimates` attributes.
        """
        pass

    def __init__(
        self,
        psms,
        score_column,
        desc=None,
        eval_fdr=0.01,
        pep_fdr_type="psm-peptide",
        prot_fdr_type="best",
        threshold=0.01,
    ):
        """Initialize a Confidence object."""
        if eval_fdr < 0 or eval_fdr > 1:
            raise ValueError("'eval_fdr' should be between 0 and 1.")
        if (
            pep_fdr_type != "psm-only"
            and pep_fdr_type != "peptide-only"
            and pep_fdr_type != "psm-peptide"
        ):
            raise ValueError(
                "'pep_fdr_type' should be 'psm-only','peptide-only', or 'psm-peptide'"
            )

        pep_fdr_type_option = ["psm-only", "peptide-only", "psm-peptide"]
        if pep_fdr_type not in pep_fdr_type_option:
            raise ValueError("%s not valid pep_fdr_type" % (pep_fdr_type))

        prot_fdr_type_option = ["best", "combine"]
        if prot_fdr_type not in prot_fdr_type_option:
            raise ValueError("%s not valid prot_fdr_type" % (prot_fdr_type))

        if desc is None:
            scores, targ = psms[score_column], psms.targets
            t_pass = (qvalues.tdc(scores, targ, desc=True) <= eval_fdr).sum()
            f_pass = (qvalues.tdc(scores, targ, desc=False) <= eval_fdr).sum()
            desc = t_pass > f_pass

        self._dataset = psms
        self._data = psms.data
        self._score_column = score_column
        self._desc = desc
        self._eval_fdr = eval_fdr
        self._levels = ("psms", "peptides", "proteins")
        self._level_columns = (
            self.dataset._spectrum_columns,
            self.dataset._peptide_column,
            self.dataset._protein_column,
        )
        self._pep_fdr_type = pep_fdr_type
        self._prot_fdr_type = prot_fdr_type
        self.confidence_estimates = {}
        self.decoy_confidence_estimates = {}

        # Assign confidence estimates
        self._assign_confidence()

        # Clean up tables
        self._prettify_tables(threshold)

    @property
    def data(self):
        """The collection of PSMs as a :py:class:`pandas.DataFrame`."""
        return self._data

    @property
    def dataset(self):
        """The underlying :py:class:`~crema.dataset.PsmDataset`"""
        return self._dataset

    @property
    def levels(self):
        """The available levels of confidence estimates"""
        return self._levels

    def _prettify_tables(self, threshold):
        """Reorder the columns of the result tables for consistency

        Parameters
        ----------
        threshold : float or "q-value", optional
            The FDR threshold for accepting discoveries. Default is 0.01. If
            "q-value" is chosen, then "accept" column is replaced with
            "crema q-value".
        """
        if threshold != "q-value":
            last_col = "accept"
        else:
            last_col = "crema q-value"

        cols = [
            *self.dataset._spectrum_columns,
            self.dataset._peptide_column,
            self.dataset._protein_column,
            self._score_column,
        ]
        prot_cols = [
            self.dataset._protein_column,
            self._score_column,
        ]
        cols.append(last_col)
        prot_cols.append(last_col)

        for level, df in self.confidence_estimates.items():
            # use 'accept' column if threshold != 'q-value'
            if threshold != "q-value":
                df[last_col] = df["crema q-value"] <= threshold

            # reverse order so best score is begining of df
            df = df.iloc[::-1]

            if level != "proteins":
                self.confidence_estimates[level] = df.loc[:, cols]
            elif level == "proteins":
                self.confidence_estimates[level] = df.loc[:, prot_cols]

        # Remove q-value column for decoy files
        cols.pop()
        prot_cols.pop()
        for level, df in self.decoy_confidence_estimates.items():
            if level != "proteins":
                self.decoy_confidence_estimates[level] = df.loc[:, cols]
            elif level == "proteins":
                self.decoy_confidence_estimates[level] = df.loc[:, prot_cols]

    def _compete(self, df, group_columns):
        """Perform target-decoy competition

        For each group defined by `group_columns`, keep only the element
        with the highest score.

        Parameters
        ----------
        df : panda.DataFrame
            The DataFrame on which to perform the competition.
        group_columns: str or list of str
            The columns that define a group. The best score is retained
            within the group.

        Returns
        -------
        pandas.DataFrame
            A :py:class:`pandas.DataFrame` containing only rows that won the
            competition.
        """
        if self._desc:
            keep = "last"
        else:
            keep = "first"

        group_columns = utils.listify(group_columns)
        # Shuffle dataframe so ties are broken randomly.
        out_df = (
            df.sample(frac=1)
            .sort_values([self._score_column] + group_columns)
            .drop_duplicates(group_columns, keep=keep)
        )

        # This ensures that best score is at top of dataframe
        if self._desc == False:
            out_df = out_df[::-1]
        return out_df

    def __getitem__(self, column):
        """Return the specified column"""
        return self._data.loc[:, column]

    def __iter__(self):
        """...but we don't want this class to be an iterable (see PEP234)"""
        raise TypeError

    def to_txt(self, output_dir=None, file_root=None, sep="\t", decoys=False):
        """Save confidence estimates to delimited text files.

        Parameters
        ----------
        output_dir : str or None, optional
            The directory in which to save the files. `None` will use the
            current working directory.
        file_root : str or None, optional
            An optional prefix for the confidence estimate files. The suffix
            will always be "crema.{level}.txt", where "{level}" indicates the
            level at which confidence estimation was performed (i.e. PSMs,
            peptides, proteins).
        sep : str, optional
            The delimiter to use.
        decoys : bool, optional
            Save decoys confidence estimates as well?

        Returns
        -------
        list of str
            The paths to the saved files.

        """
        return to_txt(
            self,
            output_dir=output_dir,
            file_root=file_root,
            sep=sep,
            decoys=decoys,
        )


class TdcConfidence(Confidence):
    """Assign confidence estimates using target decoy competition.

    Estimates q-values using the target decoy competition method. For set of
    target and decoy PSMs meeting a specified score threshold, the false
    discovery rate (FDR) is estimated as:

    .. math::
        FDR = \\frac{Decoys + 1}{Targets}

    More formally, let the scores of target and decoy PSMs be indicated as
    :math:`f_1, f_2, ..., f_{m_f}` and :math:`d_1, d_2, ..., d_{m_d}`,
    respectively. For a score threshold :math:`t`, the false discovery
    rate is estimated as:

    .. math::
        E\\{FDR(t)\\} = \\frac{|\\{d_i > t; i=1, ..., m_d\\}| + 1}
        {\\{|f_i > t; i=1, ..., m_f|\\}}

    The reported q-value for each PSM is the minimum FDR at which that
    PSM would be accepted.

    Parameters
    ----------
    psms : a PsmDataset object
        A collection of PSMs
    score_column : str, optional
        The score by which to rank the PSMs for confidence estimation. If
        :code:`None`, the score that yields the most PSMs at the specified
        false discovery rate threshold (`eval_fdr`), will be used.
    desc : bool, optional
        True if higher scores better, False if lower scores are better.
        If None, crema will try both and use the
        choice that yields the most PSMs at the specified false discovery
        rate threshold (`eval_fdr`). If `score_column` is :code:`None`,
        this parameter is ignored.
    eval_fdr : float, optional
        The false discovery rate threshold used to evaluate the best
        `score_column` and `desc` to choose. This should range from 0 to 1.
    pep_fdr_type : {"psm-only","peptide-only",psm-peptide"}, optional
        The method for crema to use when calculating peptide level confidence
        estimates.
    prot_fdr_type : {"best", "combine"}, optional
        The method for crema to use when calculating protein level confidence
        estimates. Default is "best".
    threshold : float or "q-value", optional
        The FDR threshold for accepting discoveries. Default is 0.01. If
        "q-value" is chosen, then "accept" column is replaced with
        "crema q-value".

    Attributes
    ----------
    data : pandas.DataFrame
    dataset : crema.PsmDataset
    levels : list of str
    confidence_estimates : Dict
        A dictionary containing the confidence estimates at each level, each
        as a :py:class:`pandas.DataFrame`.
    decoy_confidence_estimates : Dict
        A dictionary containing the confidence estimates for the decoy hits at
        each level, each as a :py:class:`pandas.DataFrame`
    """

    def __init__(
        self,
        psms,
        score_column=None,
        desc=None,
        eval_fdr=0.01,
        pep_fdr_type="psm-peptide",
        prot_fdr_type="best",
        threshold=0.01,
    ):
        """Initialize a TdcConfidence object."""
        LOGGER.info(
            "Assigning confidence estimates using target-decoy competition..."
        )

        super().__init__(
            psms=psms,
            score_column=score_column,
            desc=desc,
            eval_fdr=eval_fdr,
            pep_fdr_type=pep_fdr_type,
            prot_fdr_type=prot_fdr_type,
            threshold=threshold,
        )

    def _assign_confidence(self):
        """Assign confidence estimates using target-decoy competition"""
        pairing = self.dataset.peptide_pairing

        if pairing == None and self._pep_fdr_type != "psm-only":
            raise ValueError(
                "Must provide paired target decoy peptide infomation."
            )

        for level, group_cols in zip(self.levels, self._level_columns):
            # NOTE line below can removed if psm-only and peptide-only methods are removed
            df = self.data
            pair_col = utils.new_column("pairing", df)

            if level == "peptides":
                if self._pep_fdr_type == "psm-only":
                    group_cols = utils.listify(group_cols)
                elif (
                    self._pep_fdr_type == "peptide-only"
                    or self._pep_fdr_type == "psm-peptide"
                ):
                    if self._pep_fdr_type == "psm-peptide":
                        df = self._compete(df, self.dataset._spectrum_columns)
                        group_cols = utils.listify(group_cols)

                    # replace sequence with pairing
                    pair_col = utils.new_column("pairing", df)
                    df[pair_col] = df[self.dataset._peptide_column].map(
                        lambda x: pairing.get(x, x)
                    )
                    group_cols = utils.listify(group_cols) + [pair_col]
                    group_cols.remove(self.dataset._peptide_column)
                else:
                    raise ValueError(
                        f"'{self._pep_fdr_type}' is not a valid value for "
                        "pep_fdr_type "
                    )
            elif level == "proteins":
                # Perform PSM level FDR
                df = self._compete(df, self.dataset._spectrum_columns)

                # Remove peptides found in multiple proteins
                df = df[
                    ~df[self.dataset._protein_column].str.contains(
                        self.dataset._protein_delim
                    )
                ]

                # Determines how to aggregate protein score
                if self._prot_fdr_type == "best" and self._desc == True:
                    # larger score is better
                    agg_val = "max"
                elif self._prot_fdr_type == "best" and self._desc == False:
                    # smaller score is better
                    agg_val = "min"
                elif self._prot_fdr_type == "combine" and self._desc == True:
                    agg_val = "sum"
                elif self._prot_fdr_type == "combine" and self._desc == False:
                    agg_val = "prod"

                df2 = df.groupby(
                    [
                        self.dataset._protein_column,
                        self.dataset._target_column,
                    ]
                ).agg({self._score_column: [agg_val]})

                df2 = df2.reset_index()
                df2.columns = [
                    self.dataset._protein_column,
                    self.dataset._target_column,
                    self._score_column,
                ]
                df = df2

            df = self._compete(df, group_cols)
            targets = df[self.dataset._target_column]

            # Now calculate q-values:
            df["crema q-value"] = qvalues.tdc(
                scores=df[self._score_column],
                target=targets,
                desc=self._desc,
            )

            LOGGER.info(
                "  - Found %i %s at q<=%g.",
                (df[targets]["crema q-value"] <= self._eval_fdr).sum(),
                self._level_labs[level],
                self._eval_fdr,
            )

            self.confidence_estimates[level] = df.loc[targets, :]
            self.decoy_confidence_estimates[level] = df.loc[~targets, :]


class MixmaxConfidence(Confidence):
    """Assign confidence estimates using mix-max competition.

    Estimates qvalues using the mix-max competition method. To use this
    method a separate target and decoy database search using a calibrated
    score function must be used.

    # TODO Describe how mixmax works here

    Additional details can be found in this manuscript.
    U. Keich, A. Kertesz-Farkas, and W. S. Noble. Improved false
    discovery rate estimation procedure for shotgun proteomics.
    Journal of Proteome Research, 14(8):3148-3161, 2015.

    Parameters
    ----------
    psms : a PsmDataset object
        A collection of PSMs
    score_column : str, optional
        The score by which to rank the PSMs for confidence estimation. If
        :code:`None`, the score that yields the most PSMs at the specified
        false discovery rate threshold (`eval_fdr`), will be used.
    desc : bool, optional
        True if higher scores better, False if lower scores are better.
        If None, crema will try both and use the
        choice that yields the most PSMs at the specified false discovery
        rate threshold (`eval_fdr`). If `score_column` is :code:`None`,
        this parameter is ignored.
    eval_fdr : float, optional
        The false discovery rate threshold used to evaluate the best
        `score_column` and `desc` to choose. This should range from 0 to 1.
    pep_fdr_type : {"psm-only","peptide-only",psm-peptide"}, optional
        The method for crema to use when calculating peptide level confidence
        estimates. Default is "psm-peptide".
    prot_fdr_type : {"best", "combine"}, optional
        The method for crema to use when calculating protein level confidence
        estimates. Default is "best".
        estimates.
    threshold : float or "q-value", optional
        The FDR threshold for accepting discoveries. Default is 0.01. If
        "q-value" is chosen, then "accept" column is replaced with
        "crema q-value".

    Attributes
    ----------
    data : pandas.DataFrame
    dataset : crema.PsmDataset
    levels : list of str
    confidence_estimates : Dict
        A dictionary containing the confidence estimates at each level, each
        as a :py:class:`pandas.DataFrame`.
    decoy_confidence_estimates : Dict
        A dictionary containing the confidence estimates for the decoy hits at
        each level, each as a :py:class:`pandas.DataFrame`
    """

    def __init__(
        self,
        psms,
        score_column=None,
        desc=None,
        eval_fdr=0.01,
        pep_fdr_type="psm-peptide",
        prot_fdr_type="best",
        threshold=0.01,
    ):
        """Initialize a TdcConfidence object."""
        LOGGER.info(
            "Assigning confidence estimates using mix-max competition..."
        )

        super().__init__(
            psms=psms,
            score_column=score_column,
            desc=desc,
            eval_fdr=eval_fdr,
            pep_fdr_type=pep_fdr_type,
            prot_fdr_type=prot_fdr_type,
        )

    def _assign_confidence(self):
        """Assign confidence estimates using target-decoy competition"""
        # TODO maybe better way to do this
        # can not infer desc value as the wrong value will
        # result in a divide by zero error
        if self._desc == None:
            raise ValueError("'desc' has to be set for mix-max.")

        # TODO check if separate target-decoy search is done
        for level, group_cols in zip(self.levels, self._level_columns):
            if level == "peptides" or level == "proteins":
                continue

            df = self.data
            group_cols = utils.listify(group_cols)

            targets = df[df[self.dataset._target_column]]
            decoys = df[~df[self.dataset._target_column]]

            # TODO add following warning ""The mix-max procedure is not well behaved when # targets (%d) != # of decoys (%d).","

            if self._desc:
                keep = "last"
            else:
                keep = "first"

            # sort targets by score column and keep top rank
            targets_sorted = (
                targets.sample(frac=1)
                .sort_values([self._score_column] + group_cols)
                .drop_duplicates(group_cols, keep=keep, ignore_index=True)
            )

            # sort decoys by score column and keep top rank
            decoys_sorted = (
                decoys.sample(frac=1)
                .sort_values([self._score_column] + group_cols)
                .drop_duplicates(group_cols, keep=keep, ignore_index=True)
            )

            # combine top ranked target and decoy into one dataframe
            combined = pd.concat([targets_sorted, decoys_sorted])
            combined_sorted = combined.sample(frac=1).sort_values(
                [self._score_column], ascending=~self._desc, ignore_index=True
            )

            # qvalues.py::calculate_mixmax_qval expects target scores
            # and decoy scores to be sorted from worst to best
            # qvalues.py::mixmax expected combined_sorted scores
            # to be sorted from best to worst
            if self._desc:  # larger score is better
                combined_sorted = combined_sorted[::-1]
            else:  # smaller score is better
                targets_sorted[self._score_column] = (
                    targets_sorted[self._score_column] * -1.0
                )
                decoys_sorted[self._score_column] = (
                    decoys_sorted[self._score_column] * -1.0
                )
                targets_sorted = targets_sorted[::-1]
                decoys_sorted = decoys_sorted[::-1]

            # Now calculate q-values:
            pi0, targets_sorted["crema q-value"] = qvalues.mixmax(
                target_scores=targets_sorted[self._score_column],
                decoy_scores=decoys_sorted[self._score_column],
                combined_score=combined_sorted[self._score_column],
                combined_score_target=combined_sorted[
                    self.dataset._target_column
                ],
            )

            LOGGER.info("  - Estimated pi_zero = %f.", pi0)

            LOGGER.info(
                "  - Found %i %s at q<=%g.",
                (targets_sorted["crema q-value"] <= self._eval_fdr).sum(),
                self._level_labs[level],
                self._eval_fdr,
            )

            # reverse rows so that best score is at top
            targets_sorted = targets_sorted[::-1]

            # undo previous multipliation by -1.0
            if self._desc == False:
                targets_sorted[self._score_column] = (
                    targets_sorted[self._score_column] * -1.0
                )
            self.confidence_estimates[level] = targets_sorted



"""A parser for generic delmited text files."""
import logging

import pandas as pd
from ..dataset import PsmDataset
from ..utils import listify

LOGGER = logging.getLogger(__name__)


def read_txt(
    txt_files,
    target_column,
    spectrum_columns,
    score_columns,
    peptide_column,
    protein_column,
    protein_delim,
    sep="\t",
    copy_data=True,
):
    """Read peptide-spectrum matches (PSMs) from delimited text files.

    Parameters
    ----------
    txt_files : str, panda.DataFrame, or tuple of str
        One or more collection of PSMs in a tabular text format.
    target_column : str
        The column that indicates whether a PSM is a target or a decoy.
    spectrum_columns : str or tuple of str
        One or more columns that together define a unique mass spectrum.
    score_columns : str or tuple of str
        One or more columns that indicate scores by which crema can rank PSMs.
    peptide_column : str
        The column that defines a unique peptide. Modifications should be
        indicated either in square brackets :code:`[]` or parentheses
        :code:`()`. The exact modification format within these entities does
        not matter, so long as it is consistent.
    protein_column : str
        The column that defines a unique protein.
    protein_delim : str
        The delimiter to separate protein IDs.
    sep : str, optional
        The delimiter to use.
    copy_data : bool, optional
        If true, a deep copy of the data is created. This uses more memory, but
        is safer because it prevents accidental modification of the underlying
        data. This argument only has an effect when `pin_files` is a
        :py:class:`pandas.DataFrame`

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object containing the parsed
        PSMs.
    """
    # Store column names in a list to be used by read_csv function
    fields = [target_column, peptide_column, protein_column]

    # Verify some arguments are lists:
    spectrum_columns = listify(spectrum_columns)
    score_columns = listify(score_columns)
    fields += spectrum_columns + score_columns

    # Parse the data
    if isinstance(txt_files, pd.DataFrame):
        data = txt_files.copy(deep=copy_data).loc[:, fields]
    else:
        data = pd.concat(
            [_parse_psms(f, sep, fields) for f in listify(txt_files)]
        )

    data[target_column] = _convert_target_col(data[target_column])
    return PsmDataset(
        psms=data,
        target_column=target_column,
        spectrum_columns=spectrum_columns,
        score_columns=score_columns,
        peptide_column=peptide_column,
        protein_column=protein_column,
        protein_delim=protein_delim,
        copy_data=False,
    )


def _parse_psms(txt_file, sep, cols):
    """Parse a single delimited txt file.

    Parameters
    ----------
    txt_file : str
        The Crux tab-delimited file to read.
    sep : str
        The delimiter to use.
    cols : list of str
        The columns to parse.

    Returns
    -------
    pandas.DataFrame
        A :py:class:`pandas.DataFrame` containing the parsed PSMs
    """
    LOGGER.info("Reading PSMs from %s...", txt_file)
    return pd.read_csv(txt_file, sep=sep, usecols=cols)


def _convert_target_col(data):
    """Convert values in target column to boolean True/False.

    Parameters
    ----------
    data : pandas.Series
        The target column before the target/decoy column has been converted to
        boolean.

    Returns
    -------
    pandas.Series
        The target column after it has been converted to boolean.
    """
    if data.dtype == bool:
        return data
    elif data.dtype == "object":
        targets = {
            "target": True,
            "t": True,
            "decoy": False,
            "d": False,
            "f": False,
        }
        return data.map(targets)
    elif len(data.unique()) > 2:
        raise ValueError(
            "The specificed target column appears to contain more than 2 "
            "values."
        )

    return data > 0



"""A parser for the MSGF+ tab-delimited format"""
import re
import logging

import pandas as pd

from .txt import read_txt
from .. import utils

LOGGER = logging.getLogger(__name__)


def read_msgf(
    txt_files, pairing_file_name=None, decoy_prefix="XXX_", copy_data=True
):
    """Read peptide-spectrum matches (PSMs) from MSGF+ tab-delimited files.

    Parameters
    ----------
    txt_files : str, pandas.DataFrame or tuple of str
        One or more collection of PSMs in the MSGF+ tab-delimited format.
    pairing_file_name : str, optional
        A tab-delimited file that explicity pairs target and decoy peptide
        sequences. Requires one column labled 'target' that contains target
        sequences and a second colun labeled 'decoy' that contains decoy
        sequences.
    decoy_prefix : str, optional
        The prefix used to indicate a decoy protein in the protein column.
        Default value is 'XXX_'.
    copy_data : bool, optional
        If true, a deep copy of the data is created. This uses more memory, but
        is safer because it prevents accidental modification of the underlying
        data. This argument only has an effect when `txt_files` is a
        :py:class:`pandas.DataFrame`

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object containing the parsed
        PSMs.
    """
    target = "target/decoy"
    peptide = "Peptide"
    spectrum = ["#SpecFile", "ScanNum"]
    pairing = ""
    protein = "Protein"
    # TODO need to test case where protein are in diff row
    protein_delim = ";"

    # Possible score columns output by MSGF+.
    scores = {
        "DeNovoScore",
        "MSGFScore",
        "SpecEValue",
        "EValue",
    }
    scores_all = scores

    # Keep only MSGF+ scores that exist in all of the files.
    if isinstance(txt_files, pd.DataFrame):
        scores = scores.intersection(set(txt_files.columns))
    else:
        txt_files = utils.listify(txt_files)
        for txt_file in txt_files:
            with open(txt_file) as txt_ref:
                cols = txt_ref.readline().rstrip().split("\t")
                scores = scores.intersection(set(cols))

    if not scores:
        raise ValueError(
            "Could not find any of the MSGF+ score columns in all of the files."
            f"The columns Crema looks for are {', '.join(list(scores_all))}"
        )

    scores = list(scores)

    # Read in the files:
    fields = [*spectrum, peptide, target, *scores, pairing, protein]
    if isinstance(txt_files, pd.DataFrame):
        data = txt_files.copy(deep=copy_data).loc[:, fields]
    else:
        data = pd.concat(
            [utils.parse_psms_txt(f, fields, False) for f in txt_files]
        )

    data["target/decoy"] = ~data[protein].str.contains(decoy_prefix)

    psms = read_txt(
        data,
        target_column=target,
        spectrum_columns=spectrum,
        score_columns=scores,
        peptide_column=peptide,
        protein_column=protein,
        protein_delim=protein_delim,
        sep="\t",
        copy_data=False,
    )

    if pairing_file_name != None:
        psms._peptide_pairing = utils.create_pairing_from_file(
            pairing_file_name
        )

    # Remove pre/post from protein ID
    # This looks like "sp|P0AC43|SDHA_ECO57(pre=R,post=G)"
    # Remove decoy prefix from protein ID
    protein_column = psms.proteins
    new_protein_column = protein_column.str.replace(
        "\\([^()]*\\)", "", regex=True
    )
    new_protein_column = new_protein_column.str.replace(
        decoy_prefix, "", regex=True
    )
    psms.set_protein_column(new_protein_column)

    return psms



"""A parser for the MSFragger tab-delimited format"""
import re
import logging

import pandas as pd

from .txt import read_txt
from .pepxml import read_pepxml
from .pepxml import _parse_pepxml
from ..dataset import PsmDataset
from .. import utils

LOGGER = logging.getLogger(__name__)


def read_msfragger(
    txt_files, pairing_file_name=None, decoy_prefix="rev_", copy_data=True
):
    """Read peptide-spectrum matches (PSMs) from MSFragger pepXML files.

    Parameters
    ----------
    txt_files : str, pandas.DataFrame or tuple of str
        One or more collection of PSMs in the MSFragger tab-delimited format.
    pairing_file_name : str, optional
        A tab-delimited file that explicity pairs target and decoy peptide
        sequences. Requires one column labled 'target' that contains target
        sequences and a second colun labeled 'decoy' that contains decoy
        sequences.
    decoy_prefix : str, optional
        The prefix used to indicate a decoy protein in the protein column.
        Default value is 'rev_'.
    copy_data : bool, optional
        If true, a deep copy of the data is created. This uses more memory, but
        is safer because it prevents accidental modification of the underlying
        data. This argument only has an effect when `txt_files` is a
        :py:class:`pandas.DataFrame`

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object containing the parsed
        PSMs.
    """
    target = "label"
    peptide = "peptide"
    # TODO well annoying the column names for pepXML and tsv output are different
    # spectrum = ["Filename", "start scan"] # check this - is this for TSV?
    spectrum = ["ms_data_file", "scan"]
    pairing = ""
    # protein = "protein" # check for TSV
    protein = "proteins"
    protein_delim = ";"

    # The text below in any pepXML field identifies the field as a score field
    score_id = "search_engine_score:"

    # Possible score columns output by MSFragger.
    scores = {
        "hyperscore",
        "nextscore",
        "expect",
        "expectscore",
    }
    scores_all = scores

    # Read in the files:
    if isinstance(txt_files, pd.DataFrame):
        scores = scores.intersection(set(txt_files.columns))
    else:
        txt_files = utils.listify(txt_files)
        data_list = [_parse_pepxml(f, decoy_prefix) for f in txt_files]

        for data_file in data_list:
            score_col = [c for c in data_file.columns if score_id in c]
            score_col = [re.sub(score_id, "", c) for c in score_col]
            scores = scores.intersection(set(score_col))

        data = pd.concat(data_list)
        data.columns = data.columns.str.replace(score_id, "")

    if not scores:
        raise ValueError(
            "Could not find any of the MSFragger score columns in all of the "
            "files."
            f"The columns crema looks for are {', '.join(list(scores_all))}"
        )

    scores = list(scores)
    data[score_col] = data[score_col].astype(float)

    psms = PsmDataset(
        psms=data,
        target_column=target,
        spectrum_columns=spectrum,
        score_columns=scores,
        peptide_column=peptide,
        protein_column=protein,
        protein_delim=protein_delim,
        copy_data=False,
    )

    if pairing_file_name != None:
        psms._peptide_pairing = utils.create_pairing_from_file(
            pairing_file_name
        )

    return psms



"""
This is the command line interface for crema
"""
import os
import sys
import time
import logging

from .parsers.tide import read_tide
from .parsers.msamanda import read_msamanda
from .parsers.msfragger import read_msfragger
from .parsers.msgf import read_msgf
from .parsers.comet import read_comet
from .parsers.mztab import read_mztab
from .parsers.pepxml import read_pepxml
from .params import Params


def main():
    """The CLI entry point"""
    start_time = time.time()

    # Creates the parser for parse args and reads in command line arguments
    args = Params()

    # Set up logging files
    log_file = "crema.log.txt"
    if args.file_root is not None:
        log_file = args.file_root + "." + log_file
    if args.output_dir is None:
        args.output_dir = os.getcwd()

    # Configure logging
    logging.basicConfig(
        filename=os.path.join(args.output_dir, log_file),
        filemode="w+",
        level=logging.INFO,
        format="[%(levelname)s] %(message)s",
    )

    # Write logs to stderr as well
    logging.getLogger().addHandler(logging.StreamHandler())

    logging.info("crema")
    logging.info(
        "Written by Andy Lin, Donavan See and William E Fondrie in the "
    )
    logging.info(
        "Department of Genome Sciences at the University of Washington."
    )
    logging.info("Command issued:")
    logging.info("%s", " ".join(sys.argv))
    logging.info("")
    logging.info("Starting Analysis")
    logging.info("=================")

    # Create dataset object
    readers = [
        read_tide,
        read_msgf,
        read_msamanda,
        read_comet,
        read_msfragger,
        read_pepxml,
        read_mztab,
    ]

    for read_fn in readers:
        try:
            psms = read_fn(args.psm_files)
            break
        except:
            raise ValueError("Unrecognized file type.")

    conf = psms.assign_confidence(
        score_column=args.score,
        eval_fdr=args.eval_fdr,
        method=args.method,
    )

    # Write result to file
    logging.info("Writing results...")
    conf.to_txt(output_dir=args.output_dir, file_root=args.file_root)

    # Calculate how long the confidence estimation took
    end_time = time.time()
    total_time = end_time - start_time
    logging.info("==== DONE! =====")
    logging.info("Wall Time: %.2fs", total_time)


if __name__ == "__main__":
    main()



"""The :py:class:`Params` class is used to define the details and arguments
necessary for running crema from the command line.
"""
import argparse
import textwrap
from . import __version__


class CremaHelpFormatter(argparse.HelpFormatter):
    """Format help text to keep newlines and whitespace"""

    def _fill_text(self, text, width, indent):
        text_list = text.splitlines(keepends=True)
        return "\n".join(_process_line(l, width, indent) for l in text_list)


class Params:
    """
    All possible arguments and parameters for crema's CLI.
    Options can be specified as command-line arguments.
    """

    def __init__(self):
        """
        Initialize a Params object that holds an argparse parser.
        """
        self.parser = _configure_parser()
        self._namespace = vars(self.parser.parse_args())

    def __getattr__(self, option):
        return self._namespace[option]


def _configure_parser():
    """Creates and configures all the arguments for the parser"""

    desc = (
        f"crema version {__version__}\n\n"
        "Written by Andy Lin, Donavan See (seed99@cs.washington.edu), and \n"
        "William E Fondrie (wfondrie@uw.edu) in the \n"
        "Department of Genome Sciences at the University of Washington\n\n"
        "Official code website: https://github.com/Noble-Lab/crema\n\n"
        "More documentation and examples: https://crema-ms.readthedocs.io/"
    )

    parser = argparse.ArgumentParser(
        description=desc, formatter_class=CremaHelpFormatter
    )

    parser.add_argument(
        "psm_files",
        type=str,
        nargs="+",
        help=(
            "One or more collection of peptide-spectrum matches (PSMs) in the "
            "mzTab or Tide tab-delimited formats."
        ),
    )

    parser.add_argument(
        "-s",
        "--score",
        type=str,
        nargs="+",
        default=None,
        help=(
            "One or more columns that indicate possible scores by which to "
            "rank the PSMs. If more than one is provided, the best will be "
            "selected automatically. If none are provided, crema will try all "
            "available scores."
        ),
    )

    parser.add_argument(
        "-f",
        "--file_root",
        type=str,
        help="This string will be added as a prefix to all output file names.",
    )

    parser.add_argument(
        "-o",
        "--output_dir",
        type=str,
        help=(
            "The directory where output files will be created. Defaults to "
            "the current working directory."
        ),
    )

    parser.add_argument(
        "-e",
        "--eval_fdr",
        type=float,
        default=0.01,
        help=(
            "The FDR threshold by which to choose the best score column and "
            "to report in logging messages."
        ),
    )

    parser.add_argument(
        "-m",
        "--method",
        type=str,
        default="tdc",
        choices=["tdc"],
        help="The confidence estimation method to use.",
    )

    parser.add_argument(
        "-p",
        "--pep_fdr_type",
        type=str,
        default="psm-only",
        choices=["psm-only", "peptide-only", "psm-peptide"],
        help="The peptide-level FDR estimation method to use.",
    )

    parser.add_argument(
        "-r",
        "--prot_fdr_type",
        type=str,
        default="best",
        choices=["best", "combine"],
        help="The protein-level FDR estimation method to use.",
    )
    return parser


def _process_line(line, width, indent):
    """Process a line in the CLI help"""
    line = textwrap.fill(
        line,
        width,
        initial_indent=indent,
        subsequent_indent=indent,
        replace_whitespace=False,
    )
    return line.strip()



"""
These tests verify the confidence implementations function as expected
"""
import pytest
import numpy as np

from crema.confidence import TdcConfidence, MixmaxConfidence
from crema.dataset import PsmDataset

from .test_dataset import simple_df


@pytest.fixture
def simple_psms(simple_df):
    return PsmDataset(
        psms=simple_df,
        target_column="target",
        spectrum_columns=["scan", "spectrum precursor m/z"],
        score_columns=["combined p-value", "x"],
        peptide_column="sequence",
        protein_column="protein id",
        protein_delim=",",
    )


def test_tdc_confidence(simple_psms: PsmDataset):
    """Test that we can compute confidence with `TdcConfidence`"""
    conf = simple_psms.assign_confidence(
        score_column="x",
        method="tdc",
        pep_fdr_type="psm-only",
    )

    assert isinstance(conf, TdcConfidence), f"Unexpected result type: {conf}"
    # TODO: assertions


def test_mixmax_confidence(simple_psms: PsmDataset):
    """Test that we can compute confidence with `TdcConfidence`"""
    conf = simple_psms.assign_confidence(
        score_column="x",
        method="mixmax",
        pep_fdr_type="psm-only",
    )

    assert isinstance(
        conf, MixmaxConfidence
    ), f"Unexpected result type: {conf}"
    # TODO: assertions


def test_mixmax_confidence_desc(simple_psms: PsmDataset):
    """Test that we can compute confidence with `TdcConfidence`"""
    simple_psms.data["x"] = -1.0 * simple_psms.data["x"]

    conf = simple_psms.assign_confidence(
        score_column="x",
        method="mixmax",
        pep_fdr_type="psm-only",
        desc=True,
    )

    assert isinstance(
        conf, MixmaxConfidence
    ), f"Unexpected result type: {conf}"
    # TODO: assertions
